from __future__ import annotations

import ctypes

from prettyqt.qt.QtCore import Qt


user32 = ctypes.WinDLL("user32", use_last_error=True)


Mod = Qt.KeyboardModifier
Key = Qt.Key


def qt_mod_to_virtual(mod):
    vmod = 0
    mod = mod if isinstance(mod, int) else mod.value
    if mod & Mod.ShiftModifier.value:
        vmod |= MODS_TABLE.index(Mod.ShiftModifier)
    if mod & Mod.AltModifier.value:
        vmod |= MODS_TABLE.index(Mod.AltModifier)
    if mod & Mod.ControlModifier.value:
        vmod |= MODS_TABLE.index(Mod.ControlModifier)
    return vmod


def virtual_key_for_qtkey(key: int) -> int:
    old_key = key
    try:
        key = KEY_TABLE[old_key]
        if key == 0:
            key = _get_virtual_key(old_key)
    except ValueError:
        key = _get_virtual_key(old_key)
    except IndexError:
        key = KEY_TABLE.index(old_key)
        if key == 0:
            key = _get_virtual_key(old_key)
    return key


def _get_virtual_key(qtkeys):
    """Use the system keyboard layout to retrieve the virtual key.

    Fallback when we're unable to find a keycode in the mappings table.
    """
    thread_id = 0

    # Key table doesn't have an entry for this keycode
    # Attempt to retrieve the VK code from system
    keyboard_layout = user32.GetKeyboardLayout(thread_id)
    virtual_key = ctypes.windll.user32.VkKeyScanExW(qtkeys, keyboard_layout)
    if virtual_key == -1:
        keyboard_layout = user32.GetKeyboardLayout(0x409)
        virtual_key = ctypes.windll.user32.VkKeyScanExW(qtkeys, keyboard_layout)
    return virtual_key & 0xFF  # Key code is the low order byte


# https://raw.githubusercontent.com/qtproject/qtbase/dev/src/plugins/platforms/
# windows/qwindowskeymapper.cpp

# Meaning of values:
#   0 = Character output key, needs keyboard driver mapping
#   Key_unknown = Unknown Virtual Key, no translation possible, ignore

KEY_TABLE = [  # Keyboard mapping table
    # # Dec |  Hex | Windows Virtual key
    Key.Key_unknown,  # 0   0x00
    Key.Key_unknown,  # 1   0x01   VK_LBUTTON          | Left mouse button
    Key.Key_unknown,  # 2   0x02   VK_RBUTTON          | Right mouse button
    Key.Key_Cancel,  # 3   0x03   VK_CANCEL           | Control-Break processing
    Key.Key_unknown,  # 4   0x04   VK_MBUTTON          | Middle mouse button
    Key.Key_unknown,  # 5   0x05   VK_XBUTTON1         | X1 mouse button
    Key.Key_unknown,  # 6   0x06   VK_XBUTTON2         | X2 mouse button
    Key.Key_unknown,  # 7   0x07   -- unassigned --
    Key.Key_Backspace,  # 8   0x08   VK_BACK             | BackSpace key
    Key.Key_Tab,  # 9   0x09   VK_TAB              | Tab key
    Key.Key_unknown,  # 10   0x0A   -- reserved --
    Key.Key_unknown,  # 11   0x0B   -- reserved --
    Key.Key_Clear,  # 12   0x0C   VK_CLEAR            | Clear key
    Key.Key_Return,  # 13   0x0D   VK_RETURN           | Enter key
    Key.Key_unknown,  # 14   0x0E   -- unassigned --
    Key.Key_unknown,  # 15   0x0F   -- unassigned --
    Key.Key_Shift,  # 16   0x10   VK_SHIFT            | Shift key
    Key.Key_Control,  # 17   0x11   VK_CONTROL          | Ctrl key
    Key.Key_Alt,  # 18   0x12   VK_MENU             | Alt key
    Key.Key_Pause,  # 19   0x13   VK_PAUSE            | Pause key
    Key.Key_CapsLock,  # 20   0x14   VK_CAPITAL          | Caps-Lock
    Key.Key_unknown,  # 21   0x15   VK_KANA / VK_HANGUL | IME Kana or Hangul mode
    Key.Key_unknown,  # 22   0x16   -- unassigned --
    Key.Key_unknown,  # 23   0x17   VK_JUNJA            | IME Junja mode
    Key.Key_unknown,  # 24   0x18   VK_FINAL            | IME final mode
    Key.Key_unknown,  # 25   0x19   VK_HANJA / VK_KANJI | IME Hanja or Kanji mode
    Key.Key_unknown,  # 26   0x1A   -- unassigned --
    Key.Key_Escape,  # 27   0x1B   VK_ESCAPE           | Esc key
    Key.Key_unknown,  # 28   0x1C   VK_CONVERT          | IME convert
    Key.Key_unknown,  # 29   0x1D   VK_NONCONVERT       | IME non-convert
    Key.Key_unknown,  # 30   0x1E   VK_ACCEPT           | IME accept
    Key.Key_Mode_switch,  # 31   0x1F   VK_MODECHANGE       | IME mode change request
    Key.Key_Space,  # 32   0x20   VK_SPACE            | Spacebar
    Key.Key_PageUp,  # 33   0x21   VK_PRIOR            | Page Up key
    Key.Key_PageDown,  # 34   0x22   VK_NEXT             | Page Down key
    Key.Key_End,  # 35   0x23   VK_END              | End key
    Key.Key_Home,  # 36   0x24   VK_HOME             | Home key
    Key.Key_Left,  # 37   0x25   VK_LEFT             | Left arrow key
    Key.Key_Up,  # 38   0x26   VK_UP               | Up arrow key
    Key.Key_Right,  # 39   0x27   VK_RIGHT            | Right arrow key
    Key.Key_Down,  # 40   0x28   VK_DOWN             | Down arrow key
    Key.Key_Select,  # 41   0x29   VK_SELECT           | Select key
    Key.Key_Printer,  # 42   0x2A   VK_PRINT            | Print key
    Key.Key_Execute,  # 43   0x2B   VK_EXECUTE          | Execute key
    Key.Key_Print,  # 44   0x2C   VK_SNAPSHOT         | Print Screen key
    Key.Key_Insert,  # 45   0x2D   VK_INSERT           | Ins key
    Key.Key_Delete,  # 46   0x2E   VK_DELETE           | Del key
    Key.Key_Help,  # 47   0x2F   VK_HELP             | Help key
    0,  # 48   0x30   (VK_0)              | 0 key
    0,  # 49   0x31   (VK_1)              | 1 key
    0,  # 50   0x32   (VK_2)              | 2 key
    0,  # 51   0x33   (VK_3)              | 3 key
    0,  # 52   0x34   (VK_4)              | 4 key
    0,  # 53   0x35   (VK_5)              | 5 key
    0,  # 54   0x36   (VK_6)              | 6 key
    0,  # 55   0x37   (VK_7)              | 7 key
    0,  # 56   0x38   (VK_8)              | 8 key
    0,  # 57   0x39   (VK_9)              | 9 key
    Key.Key_unknown,  # 58   0x3A   -- unassigned --
    Key.Key_unknown,  # 59   0x3B   -- unassigned --
    Key.Key_unknown,  # 60   0x3C   -- unassigned --
    Key.Key_unknown,  # 61   0x3D   -- unassigned --
    Key.Key_unknown,  # 62   0x3E   -- unassigned --
    Key.Key_unknown,  # 63   0x3F   -- unassigned --
    Key.Key_unknown,  # 64   0x40   -- unassigned --
    0,  # 65   0x41   (VK_A)              | A key
    0,  # 66   0x42   (VK_B)              | B key
    0,  # 67   0x43   (VK_C)              | C key
    0,  # 68   0x44   (VK_D)              | D key
    0,  # 69   0x45   (VK_E)              | E key
    0,  # 70   0x46   (VK_F)              | F key
    0,  # 71   0x47   (VK_G)              | G key
    0,  # 72   0x48   (VK_H)              | H key
    0,  # 73   0x49   (VK_I)              | I key
    0,  # 74   0x4A   (VK_J)              | J key
    0,  # 75   0x4B   (VK_K)              | K key
    0,  # 76   0x4C   (VK_L)              | L key
    0,  # 77   0x4D   (VK_M)              | M key
    0,  # 78   0x4E   (VK_N)              | N key
    0,  # 79   0x4F   (VK_O)              | O key
    0,  # 80   0x50   (VK_P)              | P key
    0,  # 81   0x51   (VK_Q)              | Q key
    0,  # 82   0x52   (VK_R)              | R key
    0,  # 83   0x53   (VK_S)              | S key
    0,  # 84   0x54   (VK_T)              | T key
    0,  # 85   0x55   (VK_U)              | U key
    0,  # 86   0x56   (VK_V)              | V key
    0,  # 87   0x57   (VK_W)              | W key
    0,  # 88   0x58   (VK_X)              | X key
    0,  # 89   0x59   (VK_Y)              | Y key
    0,  # 90   0x5A   (VK_Z)              | Z key
    Key.Key_Meta,  # 91   0x5B   VK_LWIN             | Left Windows  - MS Natural kbd
    Key.Key_Meta,  # 92   0x5C   VK_RWIN             | Right Windows - MS Natural kbd
    Key.Key_Menu,  # 93   0x5D   VK_APPS             | Application key-MS Natural kbd
    Key.Key_unknown,  # 94   0x5E   -- reserved --
    Key.Key_Sleep,  # 95   0x5F   VK_SLEEP
    Key.Key_0,  # 96   0x60   VK_NUMPAD0          | Numeric keypad 0 key
    Key.Key_1,  # 97   0x61   VK_NUMPAD1          | Numeric keypad 1 key
    Key.Key_2,  # 98   0x62   VK_NUMPAD2          | Numeric keypad 2 key
    Key.Key_3,  # 99   0x63   VK_NUMPAD3          | Numeric keypad 3 key
    Key.Key_4,  # 100   0x64   VK_NUMPAD4          | Numeric keypad 4 key
    Key.Key_5,  # 101   0x65   VK_NUMPAD5          | Numeric keypad 5 key
    Key.Key_6,  # 102   0x66   VK_NUMPAD6          | Numeric keypad 6 key
    Key.Key_7,  # 103   0x67   VK_NUMPAD7          | Numeric keypad 7 key
    Key.Key_8,  # 104   0x68   VK_NUMPAD8          | Numeric keypad 8 key
    Key.Key_9,  # 105   0x69   VK_NUMPAD9          | Numeric keypad 9 key
    Key.Key_Asterisk,  # 106   0x6A   VK_MULTIPLY  | Multiply key
    Key.Key_Plus,  # 107   0x6B   VK_ADD           | Add key
    Key.Key_unknown,  # 108   0x6C   VK_SEPARATOR  | Separator key (locale-dependent)
    Key.Key_Minus,  # 109   0x6D   VK_SUBTRACT     | Subtract key
    Key.Key_unknown,  # 110   0x6E   VK_DECIMAL    | Decimal key (locale-dependent)
    Key.Key_Slash,  # 111   0x6F   VK_DIVIDE       | Divide key
    Key.Key_F1,  # 112   0x70   VK_F1               | F1 key
    Key.Key_F2,  # 113   0x71   VK_F2               | F2 key
    Key.Key_F3,  # 114   0x72   VK_F3               | F3 key
    Key.Key_F4,  # 115   0x73   VK_F4               | F4 key
    Key.Key_F5,  # 116   0x74   VK_F5               | F5 key
    Key.Key_F6,  # 117   0x75   VK_F6               | F6 key
    Key.Key_F7,  # 118   0x76   VK_F7               | F7 key
    Key.Key_F8,  # 119   0x77   VK_F8               | F8 key
    Key.Key_F9,  # 120   0x78   VK_F9               | F9 key
    Key.Key_F10,  # 121   0x79   VK_F10              | F10 key
    Key.Key_F11,  # 122   0x7A   VK_F11              | F11 key
    Key.Key_F12,  # 123   0x7B   VK_F12              | F12 key
    Key.Key_F13,  # 124   0x7C   VK_F13              | F13 key
    Key.Key_F14,  # 125   0x7D   VK_F14              | F14 key
    Key.Key_F15,  # 126   0x7E   VK_F15              | F15 key
    Key.Key_F16,  # 127   0x7F   VK_F16              | F16 key
    Key.Key_F17,  # 128   0x80   VK_F17              | F17 key
    Key.Key_F18,  # 129   0x81   VK_F18              | F18 key
    Key.Key_F19,  # 130   0x82   VK_F19              | F19 key
    Key.Key_F20,  # 131   0x83   VK_F20              | F20 key
    Key.Key_F21,  # 132   0x84   VK_F21              | F21 key
    Key.Key_F22,  # 133   0x85   VK_F22              | F22 key
    Key.Key_F23,  # 134   0x86   VK_F23              | F23 key
    Key.Key_F24,  # 135   0x87   VK_F24              | F24 key
    Key.Key_unknown,  # 136   0x88   -- unassigned --
    Key.Key_unknown,  # 137   0x89   -- unassigned --
    Key.Key_unknown,  # 138   0x8A   -- unassigned --
    Key.Key_unknown,  # 139   0x8B   -- unassigned --
    Key.Key_unknown,  # 140   0x8C   -- unassigned --
    Key.Key_unknown,  # 141   0x8D   -- unassigned --
    Key.Key_unknown,  # 142   0x8E   -- unassigned --
    Key.Key_unknown,  # 143   0x8F   -- unassigned --
    Key.Key_NumLock,  # 144   0x90   VK_NUMLOCK          | Num Lock key
    Key.Key_ScrollLock,  # 145   0x91   VK_SCROLL           | Scroll Lock key
    # Fujitsu/OASYS kbd --------------------
    0,  # Key.Key_Jisho, # 146   0x92   VK_OEM_FJ_JISHO     | 'Dictionary' key /
    #              VK_OEM_NEC_EQUAL  = key on numpad on NEC PC-9800 kbd
    Key.Key_Massyo,  # 147   0x93   VK_OEM_FJ_MASSHOU   | 'Unregister word' key
    Key.Key_Touroku,  # 148   0x94   VK_OEM_FJ_TOUROKU   | 'Register word' key
    0,  # Key.Key_Oyayubi_Left,#149   0x95  VK_OEM_FJ_LOYA  | 'Left OYAYUBI' key
    0,  # Key.Key_Oyayubi_Right,#150  0x96  VK_OEM_FJ_ROYA  | 'Right OYAYUBI' key
    Key.Key_unknown,  # 151   0x97   -- unassigned --
    Key.Key_unknown,  # 152   0x98   -- unassigned --
    Key.Key_unknown,  # 153   0x99   -- unassigned --
    Key.Key_unknown,  # 154   0x9A   -- unassigned --
    Key.Key_unknown,  # 155   0x9B   -- unassigned --
    Key.Key_unknown,  # 156   0x9C   -- unassigned --
    Key.Key_unknown,  # 157   0x9D   -- unassigned --
    Key.Key_unknown,  # 158   0x9E   -- unassigned --
    Key.Key_unknown,  # 159   0x9F   -- unassigned --
    Key.Key_Shift,  # 160   0xA0   VK_LSHIFT           | Left Shift key
    Key.Key_Shift,  # 161   0xA1   VK_RSHIFT           | Right Shift key
    Key.Key_Control,  # 162   0xA2   VK_LCONTROL         | Left Ctrl key
    Key.Key_Control,  # 163   0xA3   VK_RCONTROL         | Right Ctrl key
    Key.Key_Alt,  # 164   0xA4   VK_LMENU            | Left Menu key
    Key.Key_Alt,  # 165   0xA5   VK_RMENU            | Right Menu key
    Key.Key_Back,  # 166   0xA6   VK_BROWSER_BACK     | Browser Back key
    Key.Key_Forward,  # 167   0xA7   VK_BROWSER_FORWARD  | Browser Forward key
    Key.Key_Refresh,  # 168   0xA8   VK_BROWSER_REFRESH  | Browser Refresh key
    Key.Key_Stop,  # 169   0xA9   VK_BROWSER_STOP     | Browser Stop key
    Key.Key_Search,  # 170   0xAA   VK_BROWSER_SEARCH   | Browser Search key
    Key.Key_Favorites,  # 171   0xAB   VK_BROWSER_FAVORITES| Browser Favorites key
    Key.Key_HomePage,  # 172   0xAC   VK_BROWSER_HOME     | Browser Start and Home key
    Key.Key_VolumeMute,  # 173   0xAD   VK_VOLUME_MUTE      | Volume Mute key
    Key.Key_VolumeDown,  # 174   0xAE   VK_VOLUME_DOWN      | Volume Down key
    Key.Key_VolumeUp,  # 175   0xAF   VK_VOLUME_UP        | Volume Up key
    Key.Key_MediaNext,  # 176   0xB0   VK_MEDIA_NEXT_TRACK | Next Track key
    Key.Key_MediaPrevious,  # 177 0xB1   VK_MEDIA_PREV_TRACK | Previous Track key
    Key.Key_MediaStop,  # 178   0xB2   VK_MEDIA_STOP       | Stop Media key
    Key.Key_MediaTogglePlayPause,
    # 179   0xB3   VK_MEDIA_PLAY_PAUSE | Play/Pause Media key
    Key.Key_LaunchMail,  # 180   0xB4   VK_LAUNCH_MAIL      | Start Mail key
    Key.Key_LaunchMedia,  # 181   0xB5   VK_LAUNCH_MEDIA_SELECT Select Media key
    Key.Key_Launch0,  # 182   0xB6   VK_LAUNCH_APP1      | Start Application 1 key
    Key.Key_Launch1,  # 183   0xB7   VK_LAUNCH_APP2      | Start Application 2 key
    Key.Key_unknown,  # 184   0xB8   -- reserved --
    Key.Key_unknown,  # 185   0xB9   -- reserved --
    0,  # 186   0xBA   VK_OEM_1            | ';:' for US
    0,  # 187   0xBB   VK_OEM_PLUS         | '+' any country
    0,  # 188   0xBC   VK_OEM_COMMA        | ',' any country
    0,  # 189   0xBD   VK_OEM_MINUS        | '-' any country
    0,  # 190   0xBE   VK_OEM_PERIOD       | '.' any country
    0,  # 191   0xBF   VK_OEM_2            | '/?' for US
    0,  # 192   0xC0   VK_OEM_3            | '`~' for US
    Key.Key_unknown,  # 193   0xC1   -- reserved --
    Key.Key_unknown,  # 194   0xC2   -- reserved --
    Key.Key_unknown,  # 195   0xC3   -- reserved --
    Key.Key_unknown,  # 196   0xC4   -- reserved --
    Key.Key_unknown,  # 197   0xC5   -- reserved --
    Key.Key_unknown,  # 198   0xC6   -- reserved --
    Key.Key_unknown,  # 199   0xC7   -- reserved --
    Key.Key_unknown,  # 200   0xC8   -- reserved --
    Key.Key_unknown,  # 201   0xC9   -- reserved --
    Key.Key_unknown,  # 202   0xCA   -- reserved --
    Key.Key_unknown,  # 203   0xCB   -- reserved --
    Key.Key_unknown,  # 204   0xCC   -- reserved --
    Key.Key_unknown,  # 205   0xCD   -- reserved --
    Key.Key_unknown,  # 206   0xCE   -- reserved --
    Key.Key_unknown,  # 207   0xCF   -- reserved --
    Key.Key_unknown,  # 208   0xD0   -- reserved --
    Key.Key_unknown,  # 209   0xD1   -- reserved --
    Key.Key_unknown,  # 210   0xD2   -- reserved --
    Key.Key_unknown,  # 211   0xD3   -- reserved --
    Key.Key_unknown,  # 212   0xD4   -- reserved --
    Key.Key_unknown,  # 213   0xD5   -- reserved --
    Key.Key_unknown,  # 214   0xD6   -- reserved --
    Key.Key_unknown,  # 215   0xD7   -- reserved --
    Key.Key_unknown,  # 216   0xD8   -- unassigned --
    Key.Key_unknown,  # 217   0xD9   -- unassigned --
    Key.Key_unknown,  # 218   0xDA   -- unassigned --
    0,  # 219   0xDB   VK_OEM_4            | '[{' for US
    0,  # 220   0xDC   VK_OEM_5            | '\|' for US
    0,  # 221   0xDD   VK_OEM_6            | ']}' for US
    0,  # 222   0xDE   VK_OEM_7            | ''"' for US
    0,  # 223   0xDF   VK_OEM_8
    Key.Key_unknown,  # 224   0xE0   -- reserved --
    Key.Key_unknown,  # 225   0xE1   VK_OEM_AX           | 'AX' key on Japanese AX kbd
    Key.Key_unknown,  # 226   0xE2   VK_OEM_102          | "<>" or "\|" on RT 102-key kbd
    Key.Key_unknown,  # 227   0xE3   VK_ICO_HELP         | Help key on ICO
    Key.Key_unknown,  # 228   0xE4   VK_ICO_00           | 00 key on ICO
    Key.Key_unknown,  # 229   0xE5   VK_PROCESSKEY       | IME Process key
    Key.Key_unknown,  # 230   0xE6   VK_ICO_CLEAR        |
    Key.Key_unknown,  # 231   0xE7   VK_PACKET           | Unicode char as keystrokes
    Key.Key_unknown,  # 232   0xE8   -- unassigned --
    # Nokia/Ericsson definitions ---------------
    Key.Key_unknown,  # 233   0xE9   VK_OEM_RESET
    Key.Key_unknown,  # 234   0xEA   VK_OEM_JUMP
    Key.Key_unknown,  # 235   0xEB   VK_OEM_PA1
    Key.Key_unknown,  # 236   0xEC   VK_OEM_PA2
    Key.Key_unknown,  # 237   0xED   VK_OEM_PA3
    Key.Key_unknown,  # 238   0xEE   VK_OEM_WSCTRL
    Key.Key_unknown,  # 239   0xEF   VK_OEM_CUSEL
    Key.Key_unknown,  # 240   0xF0   VK_OEM_ATTN
    Key.Key_unknown,  # 241   0xF1   VK_OEM_FINISH
    Key.Key_unknown,  # 242   0xF2   VK_OEM_COPY
    Key.Key_unknown,  # 243   0xF3   VK_OEM_AUTO
    Key.Key_unknown,  # 244   0xF4   VK_OEM_ENLW
    Key.Key_unknown,  # 245   0xF5   VK_OEM_BACKTAB
    Key.Key_unknown,  # 246   0xF6   VK_ATTN             | Attn key
    Key.Key_unknown,  # 247   0xF7   VK_CRSEL            | CrSel key
    Key.Key_unknown,  # 248   0xF8   VK_EXSEL            | ExSel key
    Key.Key_unknown,  # 249   0xF9   VK_EREOF            | Erase EOF key
    Key.Key_Play,  # 250   0xFA   VK_PLAY             | Play key
    Key.Key_Zoom,  # 251   0xFB   VK_ZOOM             | Zoom key
    Key.Key_unknown,  # 252   0xFC   VK_NONAME           | Reserved
    Key.Key_unknown,  # 253   0xFD   VK_PA1              | PA1 key
    Key.Key_Clear,  # 254   0xFE   VK_OEM_CLEAR        | Clear key
    0,
]

CMD_TABLE = [  # Multimedia keys mapping table
    # Dec |  Hex | AppCommand
    Key.Key_unknown,  # 0   0x00
    Key.Key_Back,  # 1   0x01   APPCOMMAND_BROWSER_BACKWARD
    Key.Key_Forward,  # 2   0x02   APPCOMMAND_BROWSER_FORWARD
    Key.Key_Refresh,  # 3   0x03   APPCOMMAND_BROWSER_REFRESH
    Key.Key_Stop,  # 4   0x04   APPCOMMAND_BROWSER_STOP
    Key.Key_Search,  # 5   0x05   APPCOMMAND_BROWSER_SEARCH
    Key.Key_Favorites,  # 6   0x06   APPCOMMAND_BROWSER_FAVORITES
    Key.Key_Home,  # 7   0x07   APPCOMMAND_BROWSER_HOME
    Key.Key_VolumeMute,  # 8   0x08   APPCOMMAND_VOLUME_MUTE
    Key.Key_VolumeDown,  # 9   0x09   APPCOMMAND_VOLUME_DOWN
    Key.Key_VolumeUp,  # 10   0x0a   APPCOMMAND_VOLUME_UP
    Key.Key_MediaNext,  # 11   0x0b   APPCOMMAND_MEDIA_NEXTTRACK
    Key.Key_MediaPrevious,  # 12   0x0c   APPCOMMAND_MEDIA_PREVIOUSTRACK
    Key.Key_MediaStop,  # 13   0x0d   APPCOMMAND_MEDIA_STOP
    Key.Key_MediaTogglePlayPause,  # 14   0x0e   APPCOMMAND_MEDIA_PLAYPAUSE
    Key.Key_LaunchMail,  # 15   0x0f   APPCOMMAND_LAUNCH_MAIL
    Key.Key_LaunchMedia,  # 16   0x10   APPCOMMAND_LAUNCH_MEDIA_SELECT
    Key.Key_Launch0,  # 17   0x11   APPCOMMAND_LAUNCH_APP1
    Key.Key_Launch1,  # 18   0x12   APPCOMMAND_LAUNCH_APP2
    Key.Key_BassDown,  # 19   0x13   APPCOMMAND_BASS_DOWN
    Key.Key_BassBoost,  # 20   0x14   APPCOMMAND_BASS_BOOST
    Key.Key_BassUp,  # 21   0x15   APPCOMMAND_BASS_UP
    Key.Key_TrebleDown,  # 22   0x16   APPCOMMAND_TREBLE_DOWN
    Key.Key_TrebleUp,  # 23   0x17   APPCOMMAND_TREBLE_UP
    Key.Key_MicMute,  # 24   0x18   APPCOMMAND_MICROPHONE_VOLUME_MUTE
    Key.Key_MicVolumeDown,  # 25   0x19   APPCOMMAND_MICROPHONE_VOLUME_DOWN
    Key.Key_MicVolumeUp,  # 26   0x1a   APPCOMMAND_MICROPHONE_VOLUME_UP
    Key.Key_Help,  # 27   0x1b   APPCOMMAND_HELP
    Key.Key_Find,  # 28   0x1c   APPCOMMAND_FIND
    Key.Key_New,  # 29   0x1d   APPCOMMAND_NEW
    Key.Key_Open,  # 30   0x1e   APPCOMMAND_OPEN
    Key.Key_Close,  # 31   0x1f   APPCOMMAND_CLOSE
    Key.Key_Save,  # 32   0x20   APPCOMMAND_SAVE
    Key.Key_Printer,  # 33   0x21   APPCOMMAND_PRINT
    Key.Key_Undo,  # 34   0x22   APPCOMMAND_UNDO
    Key.Key_Redo,  # 35   0x23   APPCOMMAND_REDO
    Key.Key_Copy,  # 36   0x24   APPCOMMAND_COPY
    Key.Key_Cut,  # 37   0x25   APPCOMMAND_CUT
    Key.Key_Paste,  # 38   0x26   APPCOMMAND_PASTE
    Key.Key_Reply,  # 39   0x27   APPCOMMAND_REPLY_TO_MAIL
    Key.Key_MailForward,  # 40   0x28   APPCOMMAND_FORWARD_MAIL
    Key.Key_Send,  # 41   0x29   APPCOMMAND_SEND_MAIL
    Key.Key_Spell,  # 42   0x2a   APPCOMMAND_SPELL_CHECK
    Key.Key_unknown,  # 43   0x2b   APPCOMMAND_DICTATE_OR_COMMAND_CONTROL_TOGGLE
    Key.Key_unknown,  # 44   0x2c   APPCOMMAND_MIC_ON_OFF_TOGGLE
    Key.Key_unknown,  # 45   0x2d   APPCOMMAND_CORRECTION_LIST
    Key.Key_MediaPlay,  # 46   0x2e   APPCOMMAND_MEDIA_PLAY
    Key.Key_MediaPause,  # 47   0x2f   APPCOMMAND_MEDIA_PAUSE
    Key.Key_MediaRecord,  # 48   0x30   APPCOMMAND_MEDIA_RECORD
    Key.Key_AudioForward,  # 49   0x31   APPCOMMAND_MEDIA_FAST_FORWARD
    Key.Key_AudioRewind,  # 50   0x32   APPCOMMAND_MEDIA_REWIND
    Key.Key_ChannelDown,  # 51   0x33   APPCOMMAND_MEDIA_CHANNEL_DOWN
    Key.Key_ChannelUp,  # 52   0x34   APPCOMMAND_MEDIA_CHANNEL_UP
]

# Possible modifier states.
# NOTE: The order of these states match the order in
# QWindowsKeyMapper::updatePossibleKeyCodes()!
# This table is modified from original Qt version to match Modifiers in Windows, see:
# https://msdn.microsoft.com/en-us/library/windows/desktop/ms646279%28v=vs.85%29.aspx

MODS_TABLE = [
    Mod.NoModifier,  # 0
    Mod.AltModifier,  # 1
    Mod.ControlModifier,  # 2
    Mod.ControlModifier | Mod.AltModifier,  # 3
    Mod.ShiftModifier,  # 4
    Mod.ShiftModifier | Mod.AltModifier,  # 5
    Mod.ShiftModifier | Mod.ControlModifier,  # 6
    Mod.ShiftModifier | Mod.AltModifier | Mod.ControlModifier,  # 7
    Mod.NoModifier,  # Fall-back to raw Key_*
]



from __future__ import annotations

from prettyqt import widgets


class SpacerItem(widgets.LayoutItemMixin, widgets.QSpacerItem):
    """Blank space in a layout."""

    def __init__(
        self,
        w: int,
        h: int,
        h_policy: widgets.QSizePolicy.Policy
        | widgets.sizepolicy.SizePolicyStr = "minimum",
        v_policy: widgets.QSizePolicy.Policy
        | widgets.sizepolicy.SizePolicyStr = "minimum",
    ):
        if isinstance(h_policy, widgets.QSizePolicy.Policy):
            h_pol = h_policy
        else:
            h_pol = widgets.sizepolicy.SIZE_POLICY[h_policy]
        if isinstance(v_policy, widgets.QSizePolicy.Policy):
            v_pol = v_policy
        else:
            v_pol = widgets.sizepolicy.SIZE_POLICY[v_policy]
        super().__init__(w, h, h_pol, v_pol)

    def change_size(
        self,
        w: int,
        h: int,
        h_policy: widgets.sizepolicy.SizePolicyStr = "minimum",
        v_policy: widgets.sizepolicy.SizePolicyStr = "minimum",
    ):
        h_pol = widgets.sizepolicy.SIZE_POLICY[h_policy]
        v_pol = widgets.sizepolicy.SIZE_POLICY[v_policy]
        self.changeSize(w, h, h_pol, v_pol)

    def get_size_policy(self) -> widgets.SizePolicy:
        return widgets.SizePolicy.clone(self.sizePolicy())



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtWebChannel


class WebChannel(core.ObjectMixin, QtWebChannel.QWebChannel):
    pass


if __name__ == "__main__":
    channel = WebChannel()



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import constants, gui
from prettyqt.utils import colors, get_repr, serializemixin


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


class Pen(serializemixin.SerializeMixin, gui.QPen):
    """Defines how a QPainter should draw lines and outlines of shapes."""

    def __repr__(self):
        return get_repr(self, self.get_color())

    def set_color(self, color: datatypes.ColorType):
        color = colors.get_color(color)
        self.setColor(color)

    def get_color(self) -> gui.Color:
        return gui.Color(self.color())

    def get_brush(self) -> gui.Brush:
        return gui.Brush(self.brush())

    def set_cap_style(self, style: constants.CapStyleStr | constants.PenCapStyle):
        """Set cap style to use.

        Args:
            style: cap style to use
        """
        self.setCapStyle(constants.CAP_STYLE.get_enum_value(style))

    def get_cap_style(self) -> constants.CapStyleStr:
        """Return current cap style.

        Returns:
            cap style
        """
        return constants.CAP_STYLE.inverse[self.capStyle()]

    def set_join_style(self, style: constants.JoinStyleStr | constants.PenJoinStyle):
        """Set join style to use.

        Args:
            style: join style to use
        """
        self.setJoinStyle(constants.JOIN_STYLE.get_enum_value(style))

    def get_join_style(self) -> constants.JoinStyleStr:
        """Return current join style.

        Returns:
            join style
        """
        return constants.JOIN_STYLE.inverse[self.joinStyle()]

    def set_style(
        self, style: constants.PenStyleStr | constants.PenStyle | list[float] | None
    ):
        """Set pen style to use.

        Args:
            style: pen style to use
        """
        if isinstance(style, list):
            self.setDashPattern(style)
        else:
            if style is None:
                style = "none"
            self.setStyle(constants.PEN_STYLE.get_enum_value(style))

    def get_style(self) -> constants.PenStyleStr:
        """Return current pen style.

        Returns:
            pen style
        """
        return constants.PEN_STYLE.inverse[self.style()]



from __future__ import annotations

import enum
from importlib import machinery
import logging
import os
import pathlib
import pkgutil
import types
from typing import ClassVar

from prettyqt import constants, core, itemmodels


logger = logging.getLogger(__name__)


class ModuleInfoModel(itemmodels.TreeModel):
    """Tree Model to display a module hierarchy (using pkgutil)."""

    HEADER: ClassVar = ["Name", "Path", "Is Package"]
    SUPPORTS = str | os.PathLike[str] | types.ModuleType | pkgutil.ModuleInfo

    def __init__(self, obj, **kwargs):
        match obj:
            case str() | os.PathLike():
                path = pathlib.Path(obj)
                obj = pkgutil.ModuleInfo(
                    module_finder=machinery.FileFinder(str(path.parent)),
                    name=path.name,
                    ispkg=True,
                )
            case types.ModuleType():
                path = pathlib.Path(obj)
                obj = pkgutil.ModuleInfo(
                    module_finder=machinery.FileFinder(str(path.parent)),
                    name=path.name,
                    ispkg=True,
                )
            case pkgutil.ModuleInfo():
                pass
            case _:
                raise TypeError(obj)
        super().__init__(obj, **kwargs)

    @core.Enum
    class Roles(enum.IntEnum):
        ModuleInfoRole = constants.USER_ROLE + 2

    def columnCount(self, parent=None):
        return len(self.HEADER)

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> str | None:
        match orientation, role, section:
            case constants.HORIZONTAL, constants.DISPLAY_ROLE, _:
                return self.HEADER[section]
        return None

    def data(self, index: core.ModelIndex, role=constants.DISPLAY_ROLE):
        if not index.isValid():
            return None
        info = self.data_by_index(index).obj
        match role, index.column():
            case constants.DISPLAY_ROLE, 0:
                return info.name
            case constants.DISPLAY_ROLE, 1:
                return str(info.module_finder.path)
            case constants.CHECKSTATE_ROLE, 2:
                return self.to_checkstate(info.ispkg)
            case self.Roles.ModuleInfoRole:
                return info

    @classmethod
    def supports(cls, instance) -> bool:
        return isinstance(instance, pkgutil.ModuleInfo | types.ModuleType)

    def _fetch_object_children(
        self, item: ModuleInfoModel.TreeItem
    ) -> list[ModuleInfoModel.TreeItem]:
        return [
            self.TreeItem(obj=i)
            for i in pkgutil.iter_modules([
                f"{item.obj.module_finder.path}\\{item.obj.name}"
            ])
        ]

    def _has_children(self, item: ModuleInfoModel.TreeItem) -> bool:
        path = [f"{item.obj.module_finder.path}\\{item.obj.name}"]
        return any(pkgutil.iter_modules(path))


if __name__ == "__main__":
    import prettyqt
    from prettyqt import widgets

    app = widgets.app()
    view = widgets.TreeView()
    view.setRootIsDecorated(True)
    model = ModuleInfoModel(prettyqt.__path__[0], show_root=True, parent=view)
    view.set_model(model)
    view.expand_all(depth=2)
    view.resize(1000, 1000)
    view.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import core


if TYPE_CHECKING:
    from prettyqt.qt import QtLocation


class GeoCodingManager(core.ObjectMixin):
    """Support for geocoding operations."""

    def __init__(self, item: QtLocation.QGeoCodingManager):
        self.item = item

    def __getattr__(self, val):
        return getattr(self.item, val)

    def get_locale(self) -> core.Locale:
        return core.Locale(self.item.locale())



from __future__ import annotations

from typing import ClassVar, Self

from prettyqt import widgets
from prettyqt.qt import QtCore, QtGui


TYPES = {
    bool: QtCore.QMetaType.Type.Bool,
    int: QtCore.QMetaType.Type.Int,
    str: QtCore.QMetaType.Type.QString,
    float: QtCore.QMetaType.Type.Double,
    QtGui.QColor: QtCore.QMetaType.Type.QColor,
    QtGui.QCursor: QtCore.QMetaType.Type.QCursor,
    QtCore.QDate: QtCore.QMetaType.Type.QDate,
    QtCore.QSize: QtCore.QMetaType.Type.QSize,
    QtCore.QTime: QtCore.QMetaType.Type.QTime,
    list: QtCore.QMetaType.Type.QVariantList,
    QtGui.QPolygon: QtCore.QMetaType.Type.QPolygon,
    QtGui.QPolygonF: QtCore.QMetaType.Type.QPolygonF,
    QtGui.QColorSpace: QtCore.QMetaType.Type.QColorSpace,
    QtCore.QSizeF: QtCore.QMetaType.Type.QSizeF,
    QtCore.QRectF: QtCore.QMetaType.Type.QRectF,
    QtCore.QLine: QtCore.QMetaType.Type.QLine,
    QtGui.QTextLength: QtCore.QMetaType.Type.QTextLength,
    dict: QtCore.QMetaType.Type.QVariantMap,
    QtGui.QIcon: QtCore.QMetaType.Type.QIcon,
    QtGui.QPen: QtCore.QMetaType.Type.QPen,
    QtCore.QLineF: QtCore.QMetaType.Type.QLineF,
    QtGui.QTextFormat: QtCore.QMetaType.Type.QTextFormat,
    QtCore.QRect: QtCore.QMetaType.Type.QRect,
    QtCore.QPoint: QtCore.QMetaType.Type.QPoint,
    QtCore.QUrl: QtCore.QMetaType.Type.QUrl,
    QtCore.QRegularExpression: QtCore.QMetaType.Type.QRegularExpression,
    QtCore.QDateTime: QtCore.QMetaType.Type.QDateTime,
    QtCore.QPointF: QtCore.QMetaType.Type.QPointF,
    QtGui.QPalette: QtCore.QMetaType.Type.QPalette,
    QtGui.QFont: QtCore.QMetaType.Type.QFont,
    QtGui.QBrush: QtCore.QMetaType.Type.QBrush,
    QtGui.QRegion: QtCore.QMetaType.Type.QRegion,
    QtGui.QImage: QtCore.QMetaType.Type.QImage,
    QtGui.QKeySequence: QtCore.QMetaType.Type.QKeySequence,
    widgets.QSizePolicy: QtCore.QMetaType.Type.QSizePolicy,
    QtGui.QPixmap: QtCore.QMetaType.Type.QPixmap,
    QtCore.QLocale: QtCore.QMetaType.Type.QLocale,
    QtGui.QBitmap: QtCore.QMetaType.Type.QBitmap,
    QtGui.QMatrix4x4: QtCore.QMetaType.Type.QMatrix4x4,
    QtGui.QVector2D: QtCore.QMetaType.Type.QVector2D,
    QtGui.QVector3D: QtCore.QMetaType.Type.QVector3D,
    QtGui.QVector4D: QtCore.QMetaType.Type.QVector4D,
    QtGui.QQuaternion: QtCore.QMetaType.Type.QQuaternion,
    QtCore.QEasingCurve: QtCore.QMetaType.Type.QEasingCurve,
    QtCore.QJsonValue: QtCore.QMetaType.Type.QJsonValue,
    QtCore.QJsonDocument: QtCore.QMetaType.Type.QJsonDocument,
    QtCore.QModelIndex: QtCore.QMetaType.Type.QModelIndex,
    QtCore.QPersistentModelIndex: QtCore.QMetaType.Type.QPersistentModelIndex,
    QtCore.QUuid: QtCore.QMetaType.Type.QUuid,
    QtGui.QTransform: QtCore.QMetaType.Type.QTransform,
    QtCore.QByteArray: QtCore.QMetaType.Type.QByteArray,
}


def get_creator_class(editor_cls: type[widgets.QWidget], property_name: str = ""):
    class EditorCreator(widgets.ItemEditorCreatorBase):
        def createWidget(self, parent: widgets.QWidget | None) -> widgets.QWidget:
            return editor_cls(parent=parent)

        def valuePropertyName(self) -> QtCore.QByteArray:
            return QtCore.QByteArray(property_name.encode())

    return EditorCreator


class ItemEditorFactory(widgets.QItemEditorFactory):
    """Widgets for editing item data in views and delegates."""

    creators: ClassVar = []

    @classmethod
    def register_default_editor(
        cls,
        editor_cls: type[widgets.QWidget],
        typ: int | None | type = None,
        property_name: str = "",
    ):
        factory = cls.defaultFactory()
        assert factory
        creator = get_creator_class(editor_cls, property_name)()
        cls.creators.append(creator)
        match typ:
            case None:
                typ = editor_cls.staticMetaObject.userProperty().userType()  # pyright: ignore
            case type() if typ in TYPES:
                typ = TYPES[typ].value
            case int():
                pass
            case _:
                raise TypeError(typ)
        factory.registerEditor(typ, creator)  # type: ignore
        cls.setDefaultFactory(factory)

    def register_editor(
        self,
        editor_cls: type[widgets.QWidget],
        typ: int | None | type = None,
        property_name: str = "",
    ):
        creator = get_creator_class(editor_cls, property_name)()
        self.creators.append(creator)
        match typ:
            case None:
                typ = editor_cls.staticMetaObject.userProperty().userType()  # pyright: ignore
            case type() if typ in TYPES:
                typ = TYPES[typ].value
            case int():
                pass
            case _:
                raise TypeError(typ)
        self.registerEditor(typ, creator)  # type: ignore

    @classmethod
    def create_extended(cls) -> Self:
        factory = cls()
        factory.register_editor(widgets.CheckBox, bool, "")
        factory.register_editor(widgets.SpinBox, int, "value")
        factory.register_editor(widgets.LineEdit, str, "text")
        factory.register_editor(widgets.DoubleSpinBox, float, "value")
        factory.register_editor(widgets.DateEdit, QtCore.QDate, "date")
        factory.register_editor(widgets.TimeEdit, QtCore.QTime, "time")
        factory.register_editor(widgets.DateTimeEdit, QtCore.QDateTime, "dateTime")
        factory.register_editor(custom_widgets.PointEdit, QtCore.QPoint, "value")
        factory.register_editor(custom_widgets.SizeEdit, QtCore.QSize, "value")
        factory.register_editor(custom_widgets.RectEdit, QtCore.QRect, "value")
        # factory.register_editor(custom_widgets.EnumFlagWidget, 66231, "value")
        # factory.register_editor(custom_widgets.EnumComboBox, 20001, "value")
        factory.register_editor(custom_widgets.RegionEdit, QtGui.QRegion, "value")
        factory.register_editor(
            custom_widgets.SizePolicyEdit, widgets.QSizePolicy, "value"
        )
        factory.register_editor(
            widgets.KeySequenceEdit, QtGui.QKeySequence, "keySequence"
        )
        factory.register_editor(
            custom_widgets.ColorComboBox, QtGui.QColor, "current_color"
        )
        factory.register_editor(widgets.FontComboBox, QtGui.QFont, "currentFont")
        factory.register_editor(widgets.LineEdit, QtCore.QUrl, "text")
        factory.register_editor(widgets.LineEdit, QtCore.QRegularExpression, "text")
        return factory


# factory = ItemEditorFactory()
# ItemEditorFactory.setDefaultFactory(factory)

if __name__ == "__main__":
    """Run the application."""
    from prettyqt import constants, custom_widgets

    app = widgets.app()
    table_widget = widgets.TableWidget(15, 2)
    # table_widget.set_delegate(StarDelegate(), column=1)
    table_widget.setEditTriggers(
        table_widget.EditTrigger.DoubleClicked  # type: ignore
        | table_widget.EditTrigger.SelectedClicked
    )
    table_widget.set_selection_behavior("rows")
    table_widget.setHorizontalHeaderLabels(["Title", "Rating"])
    factory = ItemEditorFactory.create_extended()
    # factory = ItemEditorFactory()
    # factory.register_editor(custom_widgets.ColorChooserButton, QtGui.QColor, "color")
    factory.setDefaultFactory(factory)
    types = dict(
        color=QtGui.QColor(30, 30, 30),
        time=QtCore.QTime(1, 1, 1),
        date=QtCore.QDate(1, 1, 1),
        datetime=QtCore.QDateTime(1, 1, 1, 1, 1, 1),
        font=QtGui.QFont(),
        str="fdsf",
        int=8,
        float=8.2,
        url=QtCore.QUrl("http://www.google.de"),
        regex=QtCore.QRegularExpression("[a-z]"),
        bool=True,
        keysequence=QtGui.QKeySequence("Ctrl+A"),
        enum=widgets.AbstractItemView.EditTrigger.DoubleClicked,
    )
    for i, (k, v) in enumerate(types.items()):
        item_1 = widgets.TableWidgetItem(k)
        item_2 = widgets.TableWidgetItem()
        item_2.setData(constants.DISPLAY_ROLE, v)
        table_widget[i, 0] = item_1
        table_widget[i, 1] = item_2

    table_widget.resizeColumnsToContents()
    table_widget.resize(500, 300)
    table_widget.show()

    app.exec()



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtHelp


class HelpFilterDataMixin:
    def get_versions(self) -> list[core.VersionNumber]:
        return [core.VersionNumber(i) for i in self.versions()]

    def set_versions(self, versions: list[core.QVersionNumber]):
        self.setVersions(versions)


class HelpFilterData(HelpFilterDataMixin):
    """Details for the filters used by QHelpFilterEngine."""

    def __init__(self, item: QtHelp.QHelpFilterData):
        self.item = item

    def __getattr__(self, val):
        return getattr(self.item, val)


if __name__ == "__main__":
    data = HelpFilterData(QtHelp.QHelpFilterData())
    print(data.get_versions())



from __future__ import annotations

from typing import Literal, Self

from prettyqt.qt import QtWebEngineCore
from prettyqt.utils import bidict, datatypes


QWebEngineUrlScheme = QtWebEngineCore.QWebEngineUrlScheme

SyntaxStr = Literal["host_port_user_info", "host_port", "host", "path"]

SYNTAX: bidict[SyntaxStr, QWebEngineUrlScheme.Syntax] = bidict(
    host_port_user_info=QWebEngineUrlScheme.Syntax.HostPortAndUserInformation,
    host_port=QWebEngineUrlScheme.Syntax.HostAndPort,
    host=QWebEngineUrlScheme.Syntax.Host,
    path=QWebEngineUrlScheme.Syntax.Path,
)

FlagStr = Literal[
    "secure_scheme",
    "local_scheme",
    "local_access_allowed",
    "no_access_allowed",
    "service_workers_allowed",
    "view_source_allowed",
    "content_security_policy_ignored",
    "cors_enabled",
]

FLAGS: bidict[FlagStr, QWebEngineUrlScheme.Flag] = bidict(
    secure_scheme=QWebEngineUrlScheme.Flag.SecureScheme,
    local_scheme=QWebEngineUrlScheme.Flag.LocalScheme,
    local_access_allowed=QWebEngineUrlScheme.Flag.LocalAccessAllowed,
    no_access_allowed=QWebEngineUrlScheme.Flag.NoAccessAllowed,
    service_workers_allowed=QWebEngineUrlScheme.Flag.ServiceWorkersAllowed,
    view_source_allowed=QWebEngineUrlScheme.Flag.ViewSourceAllowed,
    content_security_policy_ignored=QWebEngineUrlScheme.Flag.ContentSecurityPolicyIgnored,
    cors_enabled=QWebEngineUrlScheme.Flag.CorsEnabled,
)


class WebEngineUrlScheme(QtWebEngineCore.QWebEngineUrlScheme):
    def get_name(self) -> str:
        return self.name().data().decode()

    @classmethod
    def get_scheme_by_name(cls, name: datatypes.ByteArrayType) -> Self:
        scheme = cls.schemeByName(datatypes.to_bytearray(name))
        return cls(scheme)

    def set_name(self, name: datatypes.ByteArrayType):
        self.setName(datatypes.to_bytearray(name))

    def set_syntax(self, syntax: SyntaxStr | QWebEngineUrlScheme.Syntax):
        """Set syntax.

        Args:
            syntax: syntax to use
        """
        self.setSyntax(SYNTAX.get_enum_value(syntax))

    def get_syntax(self) -> SyntaxStr:
        """Return syntax.

        Returns:
            syntax
        """
        return SYNTAX.inverse[self.syntax()]


if __name__ == "__main__":
    item = WebEngineUrlScheme()



from __future__ import annotations

import contextlib
import logging
from typing import TYPE_CHECKING, Any, Literal, overload

from prettyqt import constants, core
from prettyqt.utils import bidict, helpers, listdelegators, modelhelpers


if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence


logger = logging.getLogger(__name__)

CheckIndexOptionStr = Literal[
    "none", "index_is_valid", "do_not_use_parent", "parent_is_invalid"
]

CHECK_INDEX_OPTIONS: bidict[
    CheckIndexOptionStr, core.QAbstractItemModel.CheckIndexOption
] = bidict(
    none=core.QAbstractItemModel.CheckIndexOption.NoOption,
    index_is_valid=core.QAbstractItemModel.CheckIndexOption.IndexIsValid,
    do_not_use_parent=core.QAbstractItemModel.CheckIndexOption.DoNotUseParent,
    parent_is_invalid=core.QAbstractItemModel.CheckIndexOption.ParentIsInvalid,
)

LayoutChangeHintStr = Literal["none", "vertical_sort", "horizontal_sort"]

LAYOUT_CHANGE_HINT: bidict[
    LayoutChangeHintStr, core.QAbstractItemModel.LayoutChangeHint
] = bidict(
    none=core.QAbstractItemModel.LayoutChangeHint.NoLayoutChangeHint,
    vertical_sort=core.QAbstractItemModel.LayoutChangeHint.VerticalSortHint,
    horizontal_sort=core.QAbstractItemModel.LayoutChangeHint.HorizontalSortHint,
)


class AbstractItemModelMixin(core.ObjectMixin):
    IS_RECURSIVE = False  # for docs
    DELEGATE_DEFAULT = None
    DEFAULT_FLAGS = (
        constants.DRAG_ENABLED | constants.IS_ENABLED | constants.IS_SELECTABLE
    )

    def __repr__(self):
        return f"{type(self).__name__}: ({self.rowCount()}, {self.columnCount()})"

    def __len__(self) -> int:
        return self.rowCount()

    def __add__(self, other: core.QAbstractItemModel) -> core.ConcatenateTablesProxyModel:
        proxy = core.ConcatenateTablesProxyModel()
        proxy.addSourceModel(self)
        proxy.addSourceModel(other)
        return proxy

    @overload
    def __getitem__(self, index: tuple[int, int] | int) -> core.ModelIndex: ...

    @overload
    def __getitem__(
        self, index: tuple[slice, int] | tuple[int, slice] | tuple[slice, slice]
    ) -> listdelegators.ListDelegator[core.ModelIndex]: ...

    def __getitem__(
        self, index: tuple[int | slice, int | slice]
    ) -> core.ModelIndex | listdelegators.ListDelegator[core.ModelIndex]:
        # TODO: do proxies need mapToSource here?
        rowcount = self.rowCount()
        colcount = self.columnCount()
        match index:
            case int() as row, int() as col:
                if row >= rowcount or col >= rowcount:
                    raise IndexError(index)
                return self.index(row, col)
            case (row, col):
                indexes = [
                    self.index(i, j)
                    for i, j in helpers.iter_positions(row, col, rowcount, colcount)
                ]
                return listdelegators.ListDelegator(indexes)
            case int() as row:
                if row >= rowcount:
                    raise IndexError(index)
                # this here breaks PySide6 IPython test...
                return self.index(row, 0)
            case _:
                raise TypeError(index)

    def set_data(
        self,
        index: tuple[int | slice, int | slice] | core.ModelIndex,
        value: Any,
        role=constants.EDIT_ROLE,
    ):
        match index:
            case core.ModelIndex():
                self.setData(index, value, role)
            case (row, col):
                rowcount = self.rowCount()
                colcount = self.columnCount()
                for i, j in helpers.iter_positions(row, col, rowcount, colcount):
                    self.setData(self.index(i, j), value, role)
            case _:
                raise TypeError(index)

    def check_index(
        self,
        index: core.ModelIndex,
        *,
        index_is_valid: bool = False,
        do_not_use_parent: bool = False,
        parent_is_invalid: bool = False,
    ) -> bool:
        flag = core.QAbstractItemModel.CheckIndexOption.NoOption
        if index_is_valid:
            flag |= core.QAbstractItemModel.CheckIndexOption.IndexIsValid
        if do_not_use_parent:
            flag |= core.QAbstractItemModel.CheckIndexOption.DoNotUseParent
        if parent_is_invalid:
            flag |= core.QAbstractItemModel.CheckIndexOption.ParentIsInvalid
        check_flag = core.QAbstractItemModel.CheckIndexOption(0) | flag
        return self.checkIndex(index, check_flag)

    @contextlib.contextmanager
    def change_layout(self):
        """Context manager to change the layout.

        wraps calls with correct signals
        emitted at beginning: layoutAboutToBeChanged
        emitted at end: layoutChanged

        """
        self.layoutAboutToBeChanged.emit()
        yield None
        self.layoutChanged.emit()

    @contextlib.contextmanager
    def reset_model(self):
        """Context manager to reset the model.

        wraps calls with correct signals
        emitted at beginning: beginResetModel
        emitted at end: endResetModel

        """
        self.beginResetModel()
        yield None
        self.endResetModel()

    def get_column_type(
        self,
        column: int,
        *,
        rows_to_check: int = 10,
        role=constants.DISPLAY_ROLE,
    ) -> type | None:
        """Guess column data type by checking values of first rows with given role."""
        to_check = min(rows_to_check, self.rowCount())
        if to_check == 0:
            return None
        # cant combine these or make them a generator, so we do two list comps.
        indexes = [self.index(row, column) for row in range(to_check)]
        values = [self.data(i, role=role) for i in indexes]
        if all(isinstance(i, bool) for i in values):
            return bool
        if all(isinstance(i, str) for i in values):
            return str
        if all(isinstance(i, int) for i in values):
            return int
        if all(isinstance(i, float) for i in values):
            return float
        check_values = [self.data(i, role=constants.CHECKSTATE_ROLE) for i in indexes]
        return bool if None not in check_values else None

    def update_row(self, row: int):
        start_index = self.index(row, 0)
        end_index = self.index(row, self.columnCount() - 1)
        self.dataChanged.emit(start_index, end_index)

    def update_all(self):
        top_left = self.index(0, 0)
        bottom_right = self.index(self.rowCount() - 1, self.columnCount() - 1)
        self.dataChanged.emit(top_left, bottom_right)

    @contextlib.contextmanager
    def remove_row(self, row: int, parent: core.ModelIndex | None = None):
        parent = core.ModelIndex() if parent is None else parent
        self.beginRemoveRows(parent, row, row)
        yield None
        self.endRemoveRows()

    @contextlib.contextmanager
    def remove_rows(
        self,
        first: int | None = None,
        last: int | None = None,
        parent: core.ModelIndex | None = None,
    ):
        parent = core.ModelIndex() if parent is None else parent
        first = first or 0
        last = last if last is not None else self.rowCount()
        self.beginRemoveRows(parent, first, last)
        yield None
        self.endRemoveRows()

    @contextlib.contextmanager
    def remove_columns(
        self,
        first: int | None = None,
        last: int | None = None,
        parent: core.ModelIndex | None = None,
    ):
        parent = core.ModelIndex() if parent is None else parent
        first = first or 0
        last = last if last is not None else self.rowCount()
        self.beginRemoveColumns(parent, first, last)
        yield None
        self.endRemoveColumns()

    @contextlib.contextmanager
    def insert_row(self, row: int, parent: core.ModelIndex | None = None):
        parent = core.ModelIndex() if parent is None else parent
        self.beginInsertRows(parent, row, row)
        yield None
        self.endInsertRows()

    @contextlib.contextmanager
    def insert_rows(
        self,
        first: int | None = None,
        last: int | None = None,
        parent: core.ModelIndex | None = None,
    ):
        parent = core.ModelIndex() if parent is None else parent
        first = first or 0
        last = last if last is not None else self.rowCount()
        self.beginInsertRows(parent, first, last)
        yield None
        self.endInsertRows()

    @contextlib.contextmanager
    def append_rows(self, num_rows: int, parent: core.ModelIndex | None = None):
        parent = core.ModelIndex() if parent is None else parent
        row_count = self.rowCount()
        self.beginInsertRows(parent, row_count, row_count + num_rows - 1)
        yield None
        self.endInsertRows()

    @contextlib.contextmanager
    def insert_columns(
        self,
        first: int | None = None,
        last: int | None = None,
        parent: core.ModelIndex | None = None,
    ):
        parent = core.ModelIndex() if parent is None else parent
        first = first or 0
        last = last if last is not None else self.rowCount()
        self.beginInsertColumns(parent, first, last)
        yield None
        self.endInsertColumns()

    def force_reset(self):
        self.beginResetModel()
        self.endResetModel()

    def force_layoutchange(self):
        self.layoutAboutToBeChanged.emit()
        self.layoutChanged.emit()

    def get_role_names(self) -> bidict[int, str]:
        return bidict({i: v.data().decode() for i, v in self.roleNames().items()})

    def get_breadcrumbs_path(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> list[str]:
        """Get the path for the given index.

        Returns a list containing data of all indexes up to the root for given role.
        """
        pieces = [index.data(role)]
        while (index := index.parent()).isValid():
            pieces.insert(0, index.data(role))
        return pieces

    def iter_tree(
        self,
        parent_index: core.ModelIndex | None = None,
        *,
        depth: int | None = None,
        fetch_more: bool = False,
    ) -> Iterator[core.ModelIndex]:
        """Iter through all indexes of the model tree.

        Arguments:
            parent_index: parent index
            depth: maximum iteration depth
            fetch_more: call fetchMore for all indexes until canFetchMore returns False
        """
        if parent_index is None:
            # TODO: does this always equal AbstractItemView.rootIndex()?
            # parent_index = self.index(0, 0)
            parent_index = core.ModelIndex()
        if parent_index.isValid():
            yield parent_index
            if fetch_more:
                while self.canFetchMore(parent_index):
                    self.fetchMore(parent_index)
        if depth is not None and (depth := depth - 1) < 0:
            return
        for i in range(self.rowCount(parent_index)):
            child_index = self.index(i, 0, parent_index)
            yield from self.iter_tree(child_index, depth=depth, fetch_more=fetch_more)

    def search_tree(
        self,
        value: Any,
        *,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
        parent_index: core.ModelIndex | None = None,
        max_results: int | None = None,
        depth: int | None = None,
    ) -> listdelegators.ListDelegator[core.ModelIndex]:
        """Search the tree for indexes with a given value in given role.

        Compared to QAbstractItemModel.match, this method allows to set a maximum
        search depth and passing several values to search for as a list.

        Arguments:
            value: Item or list of items to search for.
            role: Index role to search in.
            parent_index: start index for searching. If None, whole tree is searched.
            max_results: stop searching after x amount of hits. 'None' means no limit.
            depth: search depth. Search depth. 'None' means no limit.
        """
        results = []
        # This makes it impossible to search for lists. I think thats fine.
        if not isinstance(value, list):
            value = [value]
        for idx in self.iter_tree(parent_index, depth=depth):
            if self.data(idx, role) in value:
                results.append(idx)
                if len(results) == max_results:
                    break
        return listdelegators.ListDelegator(results)

    def get_child_indexes(
        self, index: core.ModelIndex
    ) -> listdelegators.ListDelegator[core.ModelIndex]:
        """Get all child indexes for given index (first column only).

        To get indexes recursively, use iter_tree.

        Arguments:
            index: ModelIndex to get children from
        """
        indexes = [self.index(i, 0, index) for i in range(self.rowCount(index))]
        return listdelegators.ListDelegator(indexes)

    def get_index_key(
        self,
        index: core.ModelIndex,
        *,
        include_column: bool = False,
        parent_index: core.ModelIndex | None = None,
    ) -> tuple[tuple[int, int], ...]:
        """Return a key tuple for given ModelIndex.

        The key tuple consists either of row integers or (row, column) indices
        describing the index position from top to bottom.

        Arguments:
            index: ModelIndex to get a key for
            include_column: whether to include the column in the index key.
            parent_index: Get key up to given ModelIndex. By default, get key up to root.
        """
        key_path = []
        parent = index
        while parent.isValid() and parent != parent_index:
            key = (parent.row(), parent.column()) if include_column else parent.row()
            key_path.append(key)
            parent = parent.parent()
        return tuple(reversed(key_path))

    def index_from_key(
        self,
        key_path: Sequence[tuple[int, int] | int],
        *,
        parent_index: core.ModelIndex | None = None,
    ) -> core.ModelIndex:
        """Return a  ModelIndex for the given key path.

        Arguments:
            key_path: Key path to get an index for.
                      Should be a sequence of either (row, column)  or row indices
            parent_index: ModelIndex to start indexing from. Defaults to root index.
        """
        return modelhelpers.index_from_key(self, key_path, parent_index=parent_index)

    @staticmethod
    def to_checkstate(value: bool):
        return constants.CheckState.Checked if value else constants.CheckState.Unchecked

    def get_header_index_for_value(
        self,
        label,
        orientation: (
            constants.Orientation | constants.OrientationStr
        ) = constants.HORIZONTAL,
        *,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> int | None:
        o = constants.ORIENTATION.get_enum_value(orientation)
        count = self.columnCount() if o == constants.HORIZONTAL else self.rowCount()
        return next(
            (i for i in range(count) if self.headerData(i, o, role) == label), None
        )

    def get_table_data(
        self,
        *,
        include_index: bool = False,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
        x_range: slice | int | None = None,
        y_range: slice | int | None = None,
        parent_index: core.ModelIndex | None = None,
        use_checkstate_role: bool = False,
    ):
        parent_index = parent_index or core.ModelIndex()
        # if self.canFetchMore(parent_index):
        #     self.fetchMore(parent_index)
        match x_range:
            case None:
                colrange = range(self.columnCount(parent_index))
            case slice():
                stop = x_range.stop or self.columnCount(parent_index)
                colrange = range(x_range.start or 0, stop, x_range.step or 1)
            case int():
                colrange = range(x_range, x_range + 1)
            case _:
                raise TypeError(x_range)

        match y_range:
            case None:
                rowrange = range(self.rowCount(parent_index))
            case slice():
                stop = y_range.stop or self.rowCount(parent_index)
                rowrange = range(y_range.start or 0, stop, y_range.step or 1)
            case int():
                rowrange = range(y_range, y_range + 1)
            case _:
                raise TypeError(y_range)
        data = [
            [self.index(i, j, parent_index).data(role) for j in colrange]
            for i in rowrange
        ]
        if use_checkstate_role:
            check_data, _, __ = self.get_table_data(
                role=constants.CHECKSTATE_ROLE,
                x_range=x_range,
                y_range=y_range,
                parent_index=parent_index,
            )
            for i, row in enumerate(data):
                for j, _column in enumerate(row):
                    if check_data[i][j] is not None:
                        data[i][j] = bool(check_data[i][j])
        h_header = [self.headerData(i, constants.HORIZONTAL) for i in colrange]
        v_header = (
            [self.headerData(i, constants.VERTICAL) for i in rowrange]
            if include_index
            else None
        )
        return data, h_header, v_header

    def to_dataframe(self, **kwargs):
        import pandas as pd

        data, h_header, v_header = self.get_table_data(**kwargs)
        return pd.DataFrame(data=data, columns=h_header, index=v_header)


class AbstractItemModel(AbstractItemModelMixin, core.QAbstractItemModel):
    """The abstract interface for item model classes."""

    # @abc.abstractmethod
    # def index(self, *args, **kwargs):
    #     return NotImplemented

    # # this one is only abstract for a specific signature.
    # # @abc.abstractmethod
    # # def parent(self, *args, **kwargs):
    # #     return NotImplemented

    # @abc.abstractmethod
    # def rowCount(self, *args, **kwargs):
    #     return NotImplemented

    # @abc.abstractmethod
    # def columnCount(self, *args, **kwargs):
    #     return NotImplemented

    # @abc.abstractmethod
    # def data(self, *args, **kwargs):
    #     return NotImp


if __name__ == "__main__":

    class Test(AbstractItemModel):
        pass



from __future__ import annotations

import collections
from collections.abc import Callable
from typing import Any

from prettyqt import constants, core
from prettyqt.qt import QtGui


class AppearanceProxyModel(core.IdentityProxyModel):
    """Proxy model for changing the "style roles" of the source model.

    In contrast to
    [SliceAppearanceProxyModel](../proxymodels/sliceappearanceproxymodel.md),
    this one works in a more "static" way, meaning that you can change the
    color / font / alignment of indexes via model.setData even when the source model
    is not item-based (like a StandardItemModel).

    ### Example:
    ```py
    model = MyModel()
    proxy = itemmodels.AppearanceProxyModel()
    proxy.set_source_model(model)
    proxy.setData(proxy.index(0, 0), gui.QColor("red"), role=constants.FOREGROUND_ROLE)
    table.set_model(proxy)
    table.show()
    ```
    """

    ID = "appearance"
    ICON = "mdi.palette-outline"

    def __init__(
        self,
        foreground_default=None,
        background_default=None,
        font_default=None,
        alignment_default=None,
        **kwargs,
    ):
        self._foregrounds = collections.defaultdict(lambda: None)
        self._backgrounds = collections.defaultdict(lambda: None)
        self._alignments = collections.defaultdict(lambda: None)
        self._fonts = collections.defaultdict(lambda: None)
        self._foreground_default = foreground_default
        self._background_default = background_default
        self._font_default = font_default
        self._alignment_default = alignment_default
        super().__init__(**kwargs)

    # def setSourceModel(self, model):
    #     if (curr_model := self.sourceModel()) is not None:
    #         # curr_model.dataChanged.disconnect(self._reset)
    #         curr_model.columnsInserted.disconnect(self._reset)
    #         curr_model.columnsRemoved.disconnect(self._reset)
    #         curr_model.columnsMoved.disconnect(self._reset)

    #     with self.reset_model():
    #         super().setSourceModel(model)

    #     # model.dataChanged.connect(self._reset)
    #     model.columnsInserted.connect(self._reset)
    #     model.columnsRemoved.connect(self._reset)
    #     model.columnsMoved.connect(self._reset)

    # def _reset(self):
    #     self._foregrounds = collections.defaultdict(lambda: None)
    #     self._backgrounds = collections.defaultdict(lambda: None)
    #     self._alignments = collections.defaultdict(lambda: None)
    #     self._fonts = collections.defaultdict(lambda: None)

    def setData(
        self,
        index: core.ModelIndex,
        value: Any,
        role: constants.ItemDataRole = constants.EDIT_ROLE,
    ) -> bool:
        key = self.get_index_key(index, include_column=True)
        match role:
            case constants.FOREGROUND_ROLE:
                self._foregrounds[key] = value
                self.dataChanged.emit(index, index)
                return True
            case constants.BACKGROUND_ROLE:
                self._backgrounds[key] = value
                self.dataChanged.emit(index, index)
                return True
            case constants.FONT_ROLE:
                self._fonts[key] = value
                self.dataChanged.emit(index, index)
                return True
            case constants.ALIGNMENT_ROLE:
                self._alignments[key] = value
                self.dataChanged.emit(index, index)
                return True
            case _:
                return super().setData(index, value, role)

    def data(  # noqa: PLR0911
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        key = self.get_index_key(index, include_column=True)
        match role:
            case constants.FOREGROUND_ROLE:
                if val := self._foregrounds[key]:
                    return val
                match self._foreground_default:
                    case type() | Callable():
                        return self._foreground_default()
                    case None:
                        return super().data(index, role)
                    case _:
                        return self._foreground_default
            case constants.BACKGROUND_ROLE:
                if val := self._backgrounds[key]:
                    return val
                match self._background_default:
                    case type() | Callable():
                        return self._background_default()
                    case None:
                        return super().data(index, role)
                    case _:
                        return self._background_default
            case constants.FONT_ROLE:
                if val := self._fonts[key]:
                    return val
                match self._font_default:
                    case type() | Callable():
                        return self._font_default()
                    case None:
                        return super().data(index, role)
                    case _:
                        return self._font_default
            case constants.ALIGNMENT_ROLE:
                val = self._alignments[key]
                return val or self._alignment_default
            case _:
                return super().data(index, role)

    def set_font_default(self, font: QtGui.QFont | str):
        self._font_default = QtGui.QFont(font)
        self._fonts = collections.defaultdict(lambda: None)
        self.update_all()

    def get_font_default(self) -> QtGui.QFont:
        return self._font_default

    def set_foreground_default(self, foreground: QtGui.QColor | QtGui.QBrush | str):
        if isinstance(foreground, str):
            foreground = QtGui.QColor(foreground)
        self._foreground_default = foreground
        self._foregrounds = collections.defaultdict(lambda: None)
        self.update_all()

    def get_foreground_default(self) -> QtGui.QColor:
        return self._foreground_default

    def set_background_default(self, background: QtGui.QColor | QtGui.QBrush | str):
        if isinstance(background, str):
            background = QtGui.QColor(background)
        self._background_default = background
        self._backgrounds = collections.defaultdict(lambda: None)
        self.update_all()

    def get_background_default(self) -> QtGui.QFont:
        return self._background_default

    def set_alignment_default(
        self, alignment: constants.AlignmentFlag | constants.AlignmentStr
    ):
        if isinstance(alignment, str):
            alignment = constants.ALIGNMENTS[alignment]
        self._alignment_default = alignment
        self._alignments = collections.defaultdict(lambda: None)
        self.update_all()

    def get_alignment_default(self) -> constants.AlignmentFlag:
        return self._alignment_default

    font_default = core.Property(
        QtGui.QFont,
        get_font_default,
        set_font_default,
        doc="Default font for whole table",
    )
    foreground_default = core.Property(
        object,
        get_foreground_default,
        set_foreground_default,
        doc="Default foreground for whole table",
    )
    background_default = core.Property(
        object,
        get_background_default,
        set_background_default,
        doc="Default background for whole table",
    )
    alignment_default = core.Property(
        constants.AlignmentFlag,
        get_alignment_default,
        set_alignment_default,
        doc="Default alignment for whole table",
    )


if __name__ == "__main__":
    from prettyqt import gui, widgets

    app = widgets.app()
    table = widgets.TableView()
    table.setSortingEnabled(True)
    _source_model = core.StringListModel(["a", "b", "b", "b"])
    model = AppearanceProxyModel(parent=table)
    model.setSourceModel(_source_model)
    model.set_font_default(gui.Font.mono().family())
    model.set_background_default("green")
    index = model.index(0, 0)
    model.setData(index, QtGui.QColor("red"), constants.BACKGROUND_ROLE)
    table.set_model(model)
    table.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import constants, widgets
from prettyqt.utils import bidict


TextDirectionStr = Literal["top_to_bottom", "bottom_to_top"]

TEXT_DIRECTIONS: bidict[TextDirectionStr, widgets.QProgressBar.Direction] = bidict(
    top_to_bottom=widgets.QProgressBar.Direction.TopToBottom,
    bottom_to_top=widgets.QProgressBar.Direction.BottomToTop,
)


class ProgressBar(widgets.WidgetMixin, widgets.QProgressBar):
    """Horizontal or vertical progress bar."""

    def __init__(self, *args, text_visible: bool = True, **kwargs):
        super().__init__(*args, text_visible=text_visible, **kwargs)

    def set_alignment(self, alignment: constants.AlignmentStr | constants.AlignmentFlag):
        """Set the alignment of the layout.

        Args:
            alignment: alignment for the layout
        """
        self.setAlignment(constants.ALIGNMENTS.get_enum_value(alignment))

    def get_alignment(self) -> constants.AlignmentStr:
        """Return current alignment.

        Returns:
            alignment
        """
        return constants.ALIGNMENTS.inverse[self.alignment()]

    def set_text_direction(
        self, text_direction: TextDirectionStr | widgets.QProgressBar.Direction
    ):
        """Set the text direction of the layout.

        Args:
            text_direction: text direction for the layout
        """
        self.setTextDirection(TEXT_DIRECTIONS.get_enum_value(text_direction))

    def get_text_direction(self) -> TextDirectionStr:
        """Return current text direction.

        Returns:
            Text direction
        """
        return TEXT_DIRECTIONS.inverse[self.textDirection()]

    def set_orientation(
        self, orientation: constants.OrientationStr | constants.Orientation
    ):
        """Set the orientation of the progress bar.

        Args:
            orientation: orientation for the progress bar
        """
        self.setOrientation(constants.ORIENTATION.get_enum_value(orientation))

    def get_orientation(self) -> constants.OrientationStr:
        """Return current orientation.

        Returns:
            orientation
        """
        return constants.ORIENTATION.inverse[self.orientation()]

    def set_range(self, start: int, end: int):
        self.setRange(start, end)

    def set_value(self, value: int):
        self.setValue(value)

    def get_value(self) -> int:
        return self.value()


if __name__ == "__main__":
    app = widgets.app()
    widget = ProgressBar()
    widget.show()
    app.exec()



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from prettyqt import constants, gui, widgets


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


class SplashScreenMixin(widgets.WidgetMixin):
    def __init__(self, *args, **kwargs):
        match args:
            case (os.PathLike(), *rest):
                super().__init__(gui.Pixmap(os.fspath(args[0])), *rest, **kwargs)
            case _:
                super().__init__(*args, **kwargs)
        self.set_flags(stay_on_top=True, frameless=True)
        self.setEnabled(False)

    def __enter__(self):
        self.show()
        return self

    def __exit__(self, typ, value, traceback):
        self.hide()

    def setPixmap(self, pixmap: str | os.PathLike[str] | gui.QPixmap | None):
        match pixmap:
            case os.PathLike():
                pixmap = gui.Pixmap(os.fspath(pixmap))
            case None:
                pixmap = gui.Pixmap()
        super().setPixmap(pixmap)

    def set_pixmap_width(self, width: int):
        pixmap = self.pixmap()
        pixmap = pixmap.scaledToWidth(width)
        super().setPixmap(pixmap)

    def show_message(
        self,
        text: str,
        color: datatypes.ColorType = "black",
        h_align: constants.HorizontalAlignmentStr = "center",
        v_align: constants.VerticalAlignmentStr = "bottom",
    ):
        super().showMessage(
            text,
            color=gui.Color(color),
            alignment=constants.H_ALIGNMENT[h_align] | constants.V_ALIGNMENT[v_align],
        )


class SplashScreen(SplashScreenMixin, widgets.QSplashScreen):
    """Splash screen that can be shown during application startup."""


if __name__ == "__main__":
    app = widgets.app()
    splash = SplashScreen(gui.Pixmap())
    splash.show()
    splash.show_message("test")
    app.exec()



from __future__ import annotations

import contextlib
import logging
from typing import Any, ClassVar

from prettyqt import constants, core


logger = logging.getLogger(__name__)


class BaseFieldsModel(core.AbstractTableModel):
    HEADER: ClassVar[list[str]] = []
    DELEGATE_DEFAULT = "editor"

    def __init__(self, instance, **kwargs):
        self._instance = instance
        self._fields = self.get_fields(instance)
        self._field_names = list(self._fields.keys())
        super().__init__(**kwargs)
        self.set_instance(instance)

    def get_fields(self, instance) -> dict[str, Any]:
        return NotImplemented

    def set_instance(self, instance):
        self._instance = instance
        self._fields = self.get_fields(instance)
        self._field_names = list(self._fields.keys())
        self.update_all()

    def columnCount(self, parent=None) -> int:
        return len(self.HEADER)

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> str | None:
        match orientation, role:
            case constants.HORIZONTAL, constants.DISPLAY_ROLE:
                return self.HEADER[section]
            case constants.VERTICAL, constants.DISPLAY_ROLE:
                return self._field_names[section]

    def setData(
        self,
        index: core.ModelIndex,
        value: Any,
        role: constants.ItemDataRole = constants.EDIT_ROLE,
    ) -> bool:
        if not index.isValid():
            return None
        match role, index.column():
            case constants.USER_ROLE | constants.EDIT_ROLE, _:
                field_name = self._field_names[index.row()]
                setattr(self._instance, field_name, value)
                self.update_row(index.row())
                return True
        return False

    def rowCount(self, parent: core.ModelIndex | None = None) -> int:
        """Override for AbstractitemModel base method."""
        parent = parent or core.ModelIndex()
        if parent.column() > 0:
            return 0
        return 0 if parent.isValid() else len(self._field_names)

    def _is_writable(self, field_name: str) -> bool:
        """Dumb check, set same value and check if it works.

        Should be overridden by subclasses if possible.
        """
        val = getattr(self._instance, field_name)
        with contextlib.suppress(Exception):
            setattr(self._instance, field_name, val)
            return True
        return False

    def flags(self, index: core.ModelIndex) -> constants.ItemFlag:
        field_name = self._field_names[index.row()]
        if index.column() == 0 and self._is_writable(field_name):
            return super().flags(index) | constants.IS_EDITABLE
        return super().flags(index)



from __future__ import annotations

from prettyqt import constants, core, widgets


class ScrollBarMixin(widgets.AbstractSliderMixin):
    value_changed = core.Signal(int)

    def __init__(
        self,
        orientation: constants.Orientation | constants.OrientationStr = "horizontal",
        parent: widgets.QWidget | None = None,
    ):
        ori = constants.ORIENTATION.get_enum_value(orientation)
        super().__init__(ori, parent)
        self.valueChanged.connect(self.on_value_change)

    @classmethod
    def setup_example(cls):
        w = widgets.ScrollArea()
        w.v_scrollbar = cls(orientation="vertical", parent=w)
        return w

    def scroll_by_value(self, value: int):
        """Scroll by given distance."""
        value = min(max(self.minimum(), self.value() + value), self.maximum())
        self.setValue(value)

    def scroll_to(self, value: int):
        """Scroll to given position."""
        value = min(max(self.minimum(), value), self.maximum())
        self.setValue(value)

    def scroll_to_next_row(self):
        """Scroll to the next row."""
        self.setValue(self.value() + self.singleStep())

    def scroll_to_previous_row(self):
        """Scroll to the previous row."""
        self.setValue(self.value() - self.singleStep())

    def can_scroll(self) -> bool:
        return self.maximum() > 0


class ScrollBar(ScrollBarMixin, widgets.QScrollBar):
    """Vertical or horizontal scroll bar."""


if __name__ == "__main__":
    app = widgets.app()
    widget = widgets.PlainTextEdit("gfdgdf\n" * 1000)
    widget.show()
    app.exec()



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Literal

from prettyqt import core
from prettyqt.qt import QtGui
from prettyqt.utils import bidict


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


mod = QtGui.QAbstractFileIconProvider

ICON_TYPE = bidict(
    computer=mod.IconType.Computer,
    desktop=mod.IconType.Desktop,
    trashcan=mod.IconType.Trashcan,
    network=mod.IconType.Network,
    drive=mod.IconType.Drive,
    folder=mod.IconType.Folder,
    file=mod.IconType.File,
)

IconTypeStr = Literal[
    "computer", "desktop", "trashcan", "network", "drive", "folder", "file"
]

OPTIONS = bidict(
    dont_use_custom_dir_icons=mod.Option.DontUseCustomDirectoryIcons,
)

OptionStr = Literal["dont_use_custom_dir_icons"]


class AbstractFileIconProviderMixin:
    def get_type(self, file: core.QFileInfo | datatypes.PathType) -> str:
        if isinstance(file, os.PathLike):
            file = core.FileInfo(file)
        return self.type(file)


class AbstractFileIconProvider(
    AbstractFileIconProviderMixin, QtGui.QAbstractFileIconProvider
):
    """File icons for the QFileSystemModel class."""



from __future__ import annotations

from collections.abc import Iterable

from prettyqt import constants, core, widgets


class TreeViewMixin(widgets.AbstractItemViewMixin):
    def __init__(
        self,
        *args,
        root_is_decorated: bool = True,
        all_columns_show_focus: bool = True,
        uniform_row_heights: bool = True,
        alternating_row_colors: bool = True,
        selection_mode: str = "extended",
        **kwargs,
    ):
        super().__init__(
            *args,
            root_is_decorated=root_is_decorated,
            all_columns_show_focus=all_columns_show_focus,
            uniform_row_heights=uniform_row_heights,
            alternating_row_colors=alternating_row_colors,
            selection_mode=selection_mode,
            **kwargs,
        )
        self.setStyleSheet("QTreeView::item { padding: 5px }")
        self.h_header = widgets.HeaderView("horizontal", parent=self)

    @property
    def h_header(self) -> widgets.HeaderView:
        return self.header()

    @h_header.setter
    def h_header(self, header: widgets.QHeaderView):
        self.setHeader(header)

    def show_root(self, value: bool):
        self.setRootIndex(core.QModelIndex())
        if not value:
            self.setRootIndex(self.model().index(0, 0))

    def set_expanded(
        self, index: core.ModelIndex | Iterable[core.ModelIndex], expanded: bool = True
    ):
        """Set expaned state of an index or an Iterable of indexes."""
        match index:
            case Iterable():
                for idx in index:
                    self.setExpanded(idx, expanded)
            case _:
                self.setExpanded(index, expanded)

    def expand_all(self, depth: int | None = None):
        if depth is None:
            super().expandAll()
        else:
            for idx in self.model().iter_tree(parent_index=self.rootIndex(), depth=depth):
                self.setExpanded(idx, True)

    def set_sorting_enabled(self, enabled: bool, do_sort: bool = False):
        """Hack to avoid direct sort when setting sorting enabled."""
        model = self.model()
        if not do_sort and model is not None:
            backup = model.sort
            model.sort = lambda x, y: None
        self.setSortingEnabled(enabled)
        if not do_sort and model is not None:
            model.sort = backup

    def set_indentation(self, indentation: int):
        self.setIndentation(indentation)

    def setup_list_style(self):
        self.setSelectionBehavior(self.SelectionBehavior.SelectRows)
        self.h_header.setStretchLastSection(True)

    def adapt_sizes(self):
        model = self.model()
        if model is not None and (model.rowCount() * model.columnCount()) < 1000:  # noqa: PLR2004
            self.h_header.resizeSections(self.h_header.ResizeMode.ResizeToContents)
        else:
            self.h_header.resize_sections("interactive")

    def sort_by_column(self, column: int | None, ascending: bool = True):
        column = -1 if column is None else column
        order = constants.ASCENDING if ascending else constants.DESCENDING
        self.sortByColumn(column, order)


class TreeView(TreeViewMixin, widgets.QTreeView):
    """Default model/view implementation of a tree view."""


if __name__ == "__main__":
    app = widgets.app()
    dlg = widgets.MainWindow()
    status_bar = TreeView()
    dlg.show()
    app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import constants, core
from prettyqt.utils import bidict, datatypes, get_repr


NameTypeStr = Literal["default", "long", "short", "offset"]

NAME_TYPE: bidict[NameTypeStr, core.QTimeZone.NameType] = bidict(
    default=core.QTimeZone.NameType.DefaultName,
    long=core.QTimeZone.NameType.LongName,
    short=core.QTimeZone.NameType.ShortName,
    offset=core.QTimeZone.NameType.OffsetName,
)

TimeTypeStr = Literal["standard", "daylight", "generic"]

TIME_TYPE: bidict[TimeTypeStr, core.QTimeZone.TimeType] = bidict(
    standard=core.QTimeZone.TimeType.StandardTime,
    daylight=core.QTimeZone.TimeType.DaylightTime,
    generic=core.QTimeZone.TimeType.GenericTime,
)


class TimeZone(core.QTimeZone):
    """Identifies how a time representation relates to UTC."""

    def __init__(self, *args):
        match args:
            case (str() as string, *rest):
                super().__init__(core.QByteArray(string.encode()), *rest)
            case _:
                super().__init__(*args)

    def __repr__(self):
        return get_repr(self, self.get_id())

    def __str__(self):
        return self.get_id()

    def __reduce__(self):
        return type(self), (self.get_id(),)

    def get_id(self) -> str:
        return self.id().data().decode()

    def get_display_name(
        self,
        date_time: datatypes.DateTimeType | TimeTypeStr,
        name_type: NameTypeStr | core.QTimeZone.NameType = "default",
        locale: core.QLocale | None = None,
    ) -> str:
        if date_time in list(TIME_TYPE) + list(TIME_TYPE.inverse):  # needs rework
            dt = TIME_TYPE.get_enum_value(date_time)
        else:
            dt = datatypes.to_datetime(date_time)
        return self.displayName(
            dt,
            NAME_TYPE.get_enum_value(name_type),
            locale or core.QLocale(),
        )

    def get_time_spec(self) -> constants.TimeSpecStr:
        return constants.TIME_SPEC.inverse[self.timeSpec()]

    # def get_value(self) -> datetime.datetime:
    #     try:
    #         return self.toPython()
    #     except TypeError:
    #         return self.toPyTimeZone()


if __name__ == "__main__":
    date = core.TimeZone(2000)
    dt = TimeZone(date)
    print(dt.get_display_name("standard", "short"))



from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING, Literal
import webbrowser

from prettyqt import core, gui, webenginecore
from prettyqt.utils import bidict, datatypes


if TYPE_CHECKING:
    from collections.abc import Callable


mod = webenginecore.QWebEnginePage

logger = logging.getLogger(__name__)


FeatureStr = Literal[
    "notifications",
    "geolocation",
    "media_audio_capture",
    "media_video_capture",
    "media_audiovideo_capture",
    "mouse_lock",
    "desktop_video_capture",
    "desktop_audiovideo_capture",
]

FEATURE: bidict[FeatureStr, mod.Feature] = bidict(
    notifications=mod.Feature.Notifications,
    geolocation=mod.Feature.Geolocation,
    media_audio_capture=mod.Feature.MediaAudioCapture,
    media_video_capture=mod.Feature.MediaVideoCapture,
    media_audiovideo_capture=mod.Feature.MediaAudioVideoCapture,
    mouse_lock=mod.Feature.MouseLock,
    desktop_video_capture=mod.Feature.DesktopVideoCapture,
    desktop_audiovideo_capture=mod.Feature.DesktopAudioVideoCapture,
)

FileSelectionModeStr = Literal["open", "open_multiple"]

FILE_SELECTION_MODE: bidict[FileSelectionModeStr, mod.FileSelectionMode] = bidict(
    open=mod.FileSelectionMode.FileSelectOpen,
    open_multiple=mod.FileSelectionMode.FileSelectOpenMultiple,
)

FindFlagStr = Literal["backward", "case_sensitive"]

FIND_FLAGS: bidict[FindFlagStr, mod.FindFlag] = bidict(
    backward=mod.FindFlag.FindBackward,
    case_sensitive=mod.FindFlag.FindCaseSensitively,
)

JavaScriptConsoleMessageLevelStr = Literal["info", "warning", "error"]

JS_CONSOLE_MESSAGE_LEVEL: bidict[
    JavaScriptConsoleMessageLevelStr, mod.JavaScriptConsoleMessageLevel
] = bidict(
    info=mod.JavaScriptConsoleMessageLevel.InfoMessageLevel,
    warning=mod.JavaScriptConsoleMessageLevel.WarningMessageLevel,
    error=mod.JavaScriptConsoleMessageLevel.ErrorMessageLevel,
)

LifecycleStateStr = Literal["active", "frozen", "discarded"]

LIFECYCLE_STATE: bidict[LifecycleStateStr, mod.LifecycleState] = bidict(
    active=mod.LifecycleState.Active,
    frozen=mod.LifecycleState.Frozen,
    discarded=mod.LifecycleState.Discarded,
)

NAVIGATION_TYPES = bidict(
    link_clicked=mod.NavigationType.NavigationTypeLinkClicked,
    typed=mod.NavigationType.NavigationTypeTyped,
    form_submitted=mod.NavigationType.NavigationTypeFormSubmitted,
    back_forward=mod.NavigationType.NavigationTypeBackForward,
    reload=mod.NavigationType.NavigationTypeReload,
    redirect=mod.NavigationType.NavigationTypeRedirect,
    other=mod.NavigationType.NavigationTypeOther,
)

PermissionPolicyStr = Literal["unknown", "granted_by_user", "denied_by_user"]

PERMISSION_POLICY: bidict[PermissionPolicyStr, mod.PermissionPolicy] = bidict(
    unknown=mod.PermissionPolicy.PermissionUnknown,
    granted_by_user=mod.PermissionPolicy.PermissionGrantedByUser,
    denied_by_user=mod.PermissionPolicy.PermissionDeniedByUser,
)

RENDER_PROCESS_TERMINATION_STATUS = bidict(
    normal=mod.RenderProcessTerminationStatus.NormalTerminationStatus,
    abnormal=mod.RenderProcessTerminationStatus.AbnormalTerminationStatus,
    crashed=mod.RenderProcessTerminationStatus.CrashedTerminationStatus,
    killed=mod.RenderProcessTerminationStatus.KilledTerminationStatus,
)

WebActionStr = Literal[
    "none",
    "back",
    "forward",
    "stop",
    "reload",
    "reload_and_bypass_cache",
    "cut",
    "copy",
    "paste",
    "undo",
    "redo",
    "select_all",
    "paste_and_match_style",
    "open_link_in_this_window",
    "open_link_in_new_window",
    "open_link_in_new_tab",
    "open_link_in_new_bg_tab",
    "copy_link_to_clipboard",
    "copy_image_to_clipboard",
    "copy_image_url_to_clipboard",
    "copy_media_url_to_clipboard",
    "toggle_media_controls",
    "toggle_media_loop",
    "toggle_media_play_pause",
    "toggle_media_mute",
    "download_link_to_disk",
    "download_image_to_disk",
    "download_media_to_disk",
    "inspect_element",
    "exit_fullscreen",
    "request_close",
    "unselect",
    "save_page",
    "view_source",
    "toggle_bold",
    "toggle_italic",
    "toggle_underline",
    "toggle_strikethrough",
    "align_left",
    "align_center",
    "align_right",
    "align_justified",
    "indent",
    "outdent",
    "insert_ordered_list",
    "insert_unordered_list",
]

WEB_ACTION: bidict[WebActionStr, mod.WebAction] = bidict(
    none=mod.WebAction.NoWebAction,
    back=mod.WebAction.Back,
    forward=mod.WebAction.Forward,
    stop=mod.WebAction.Stop,
    reload=mod.WebAction.Reload,
    reload_and_bypass_cache=mod.WebAction.ReloadAndBypassCache,
    cut=mod.WebAction.Cut,
    copy=mod.WebAction.Copy,
    paste=mod.WebAction.Paste,
    undo=mod.WebAction.Undo,
    redo=mod.WebAction.Redo,
    select_all=mod.WebAction.SelectAll,
    paste_and_match_style=mod.WebAction.PasteAndMatchStyle,
    open_link_in_this_window=mod.WebAction.OpenLinkInThisWindow,
    open_link_in_new_window=mod.WebAction.OpenLinkInNewWindow,
    open_link_in_new_tab=mod.WebAction.OpenLinkInNewTab,
    open_link_in_new_bg_tab=mod.WebAction.OpenLinkInNewBackgroundTab,
    copy_link_to_clipboard=mod.WebAction.CopyLinkToClipboard,
    copy_image_to_clipboard=mod.WebAction.CopyImageToClipboard,
    copy_image_url_to_clipboard=mod.WebAction.CopyImageUrlToClipboard,
    copy_media_url_to_clipboard=mod.WebAction.CopyMediaUrlToClipboard,
    toggle_media_controls=mod.WebAction.ToggleMediaControls,
    toggle_media_loop=mod.WebAction.ToggleMediaLoop,
    toggle_media_play_pause=mod.WebAction.ToggleMediaPlayPause,
    toggle_media_mute=mod.WebAction.ToggleMediaMute,
    download_link_to_disk=mod.WebAction.DownloadLinkToDisk,
    download_image_to_disk=mod.WebAction.DownloadImageToDisk,
    download_media_to_disk=mod.WebAction.DownloadMediaToDisk,
    inspect_element=mod.WebAction.InspectElement,
    exit_fullscreen=mod.WebAction.ExitFullScreen,
    request_close=mod.WebAction.RequestClose,
    unselect=mod.WebAction.Unselect,
    save_page=mod.WebAction.SavePage,
    view_source=mod.WebAction.ViewSource,
    toggle_bold=mod.WebAction.ToggleBold,
    toggle_italic=mod.WebAction.ToggleItalic,
    toggle_underline=mod.WebAction.ToggleUnderline,
    toggle_strikethrough=mod.WebAction.ToggleStrikethrough,
    align_left=mod.WebAction.AlignLeft,
    align_center=mod.WebAction.AlignCenter,
    align_right=mod.WebAction.AlignRight,
    align_justified=mod.WebAction.AlignJustified,
    indent=mod.WebAction.Indent,
    outdent=mod.WebAction.Outdent,
    insert_ordered_list=mod.WebAction.InsertOrderedList,
    insert_unordered_list=mod.WebAction.InsertUnorderedList,
)

WEB_WINDOW_TYPES = bidict(
    browser_window=mod.WebWindowType.WebBrowserWindow,
    browser_tab=mod.WebWindowType.WebBrowserTab,
    dialog=mod.WebWindowType.WebDialog,
    browser_bg_tab=mod.WebWindowType.WebBrowserBackgroundTab,
)


class WebEnginePage(core.ObjectMixin, webenginecore.QWebEnginePage):
    """A web engine page holds the HTML document contents, link history + actions."""

    def get_icon(self) -> gui.Icon | None:
        """Return icon. If icon is Null, return None."""
        icon = self.icon()
        return None if icon.isNull() else gui.Icon(icon)

    def set_url(self, url: datatypes.PathType | datatypes.UrlType):
        """Set the url of the WebEnginePage.

        Clears the Page and loads the URL.

        Args:
            url: URL to set
        """
        self.setUrl(datatypes.to_url(url))

    def get_url(self) -> core.Url:
        return core.Url(self.url())

    def get_icon_url(self) -> core.Url:
        return core.Url(self.iconUrl())

    def get_requested_url(self) -> core.Url:
        return core.Url(self.requestedUrl())

    def get_scroll_position(self) -> core.PointF:
        return core.PointF(self.scrollPosition())

    def get_contents_size(self) -> core.SizeF:
        return core.SizeF(self.contentsSize())

    def load_url(self, url: datatypes.UrlType | datatypes.PathType):
        """Load the URL.

        Loads the specified url and displays it.

        Note: The Page remains the same until enough data has arrived
        to display the new URL.

        Args:
            url: URL to load
        """
        self.load(datatypes.to_url(url))

    def set_zoom(self, zoom: float):
        """Set the zoom factor for the Page.

        Valid values are within the range from 0.25 to 5.0. The default factor is 1.0.

        Args:
            zoom: Zoom factor
        """
        self.setZoomFactor(zoom)

    def find_text(
        self,
        string: str,
        backward: bool = False,
        case_sensitive: bool = False,
        callback: Callable[[bool], None] | None = None,
    ):
        """Find text in the current page.

        Finds the specified string, subString, in the page, using the given options.
        The findTextFinished() signal is emitted when a string search is completed.

        To clear the search highlight, just pass an empty string.

        The resultCallback must take a boolean parameter.
        It will be called with a value of true if the subString was found;
        otherwise the callback value will be false.

        Warning: It is guaranteed that the callback is always called,
        but it might be done during page destruction. When WebEnginePage is deleted,
        the callback is triggered with an invalid value and it is not safe to use
        the corresponding QWebEnginePage or QWebEnginePage instance inside it.

        Args:
            string: string to search for
            backward: search backwards
            case_sensitive: case-sensitive search
            callback: result callback
        """
        if callback is None:

            def do_nothing(x):
                pass

            callback = do_nothing
        flag = webenginecore.QWebEnginePage.FindFlag(0)
        if case_sensitive:
            flag |= self.FindFlag.FindCaseSensitively
        if backward:
            flag |= self.FindFlag.FindBackward
        self.findText(string, flag, callback)

    def set_lifecycle_state(self, state: LifecycleStateStr | mod.LifecycleState):
        """Set lifecycle state.

        Args:
            state: lifecycle state
        """
        self.setLifecycleState(LIFECYCLE_STATE.get_enum_value(state))

    def get_lifecycle_state(self) -> LifecycleStateStr:
        """Get the current lifecycle state.

        Returns:
            lifecycle state
        """
        return LIFECYCLE_STATE.inverse[self.lifecycleState()]

    def trigger_action(self, action: WebActionStr | mod.WebAction, checked: bool = False):
        """Trigger action."""
        self.triggerAction(WEB_ACTION.get_enum_value(action), checked)

    def set_feature_permission(
        self,
        url: datatypes.UrlType,
        feature: FeatureStr | mod.Feature,
        policy: PermissionPolicyStr | mod.PermissionPolicy,
    ):
        """Set permission of feature for given URL."""
        url = core.Url(url)
        self.setFeaturePermission(
            url, FEATURE.get_enum_value(feature), PERMISSION_POLICY.get_enum_value(policy)
        )

    def get_history(self) -> webenginecore.WebEngineHistory:
        hist = self.history()
        return webenginecore.WebEngineHistory(hist)

    def get_settings(self) -> webenginecore.WebEngineSettings:
        """Get WebEngineSettings (a MutableMapping)."""
        settings = self.settings()
        return webenginecore.WebEngineSettings(settings)

    def set_setting(
        self,
        setting_name: webenginecore.webenginesettings.WebAttributeStr,
        value: bool,
    ):
        """Set WebEngine setting to value."""
        self.get_settings()[setting_name] = value

    def get_setting(
        self,
        setting_name: webenginecore.webenginesettings.WebAttributeStr,
    ) -> bool:
        """Return value of given WebEngine setting."""
        return self.get_settings()[setting_name]

    def get_scripts(self) -> webenginecore.WebEngineScriptCollection:
        """Get script collection."""
        return webenginecore.WebEngineScriptCollection(self.scripts())

    def open_in_browser(self):
        """Open page URL in default system browser."""
        try:
            webbrowser.open(self.getUrl().toString())
        except ValueError:
            logger.exception("Error opening URL %r", self.getUrl().toString())

    def insert_stylesheet(
        self, name: str, css: str | os.PathLike[str], immediately: bool = True
    ):
        """Load css using JavaScript.

        Arguments:
            name: CSS id
            css: CSS stylesheet, either a string or a PathLike object.
            immediately: whther to run javascript immediately
        """
        if isinstance(css, os.PathLike):
            path = core.File(os.fspath(css))
            if not path.open(
                core.File.OpenModeFlag.ReadOnly | core.File.OpenModeFlag.Text
            ):
                return
            css = path.readAll().data().decode("utf-8")
        script = webenginecore.WebEngineScript()
        s = """
        (function() {
        css = document.createElement('style');
        css.type = 'text/css';
        css.id = "%s";
        document.head.appendChild(css);
        css.innerText = `%s`;
        })()
        """
        s = s % (name, css)
        script = webenginecore.WebEngineScript()
        if immediately:
            self.runJavaScript(s, script.ScriptWorldId.ApplicationWorld)
        script.setName(name)
        script.setSourceCode(s)
        script.setInjectionPoint(script.InjectionPoint.DocumentReady)
        script.setRunsOnSubFrames(True)
        script.setWorldId(script.ScriptWorldId.ApplicationWorld)
        self.scripts().insert(script)

    def mousedown(self, selector: str, btn: Literal["left", "middle", "right"]):
        """Simulate a mousedown event on the targeted element.

        Arguments:
            selector: A CSS3 selector to targeted element
            btn: Mouse button
        """
        btn_id = dict(left=0, middle=1, right=2)[btn]
        return self.runJavaScript(
            f"""
            (function () {{
                var element = document.querySelector({selector!r});
                var evt = document.createEvent("MouseEvents");
                evt.initMouseEvent("mousedown", true, true, window,
                                   1, 1, 1, 1, 1, false, false, false, false,
                                   {btn_id!r}, element);
                return element.dispatchEvent(evt);
            }})();
        """
        )

    def set_input_value(self, selector: str, value):
        """Set the value of the input matched by given selector."""
        script = f'document.querySelector({selector!r}).setAttribute("value", "{value}")'
        self.runJavaScript(script)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    path = path = "E:\\dev\\datacook\\processanalyzer\\docs\\index.html"
    widget = WebEnginePage()
    widget.set_url(path)
    widget.find_text("test", backward=True, case_sensitive=True, callback=None)
    app.exec()



from __future__ import annotations

from prettyqt import constants, core, gui, widgets


class IconDelegate(widgets.StyledItemDelegate):
    """Delegate to paint QIcons, QPixmaps, QColors and QImages."""

    ID = "icon"

    def __init__(self, role: constants.ItemDataRole = constants.USER_ROLE, **kwargs):
        self._role = role
        self.margin = 10
        super().__init__(**kwargs)

    def paint(
        self,
        painter: gui.QPainter,
        option: widgets.QStyleOptionViewItem,
        index: core.ModelIndex,
    ):
        """Override to paint an icon based on given Pixmap / Color / Icon.

        Pixmap / Color / Icon must be set to '_role'

        Args:
            painter (gui.QPainter): painter to paint the icon
            option (widgets.QStyleOptionViewItem): state of the item to be displayed
            index (core.ModelIndex): index which gets decorated
        """
        super().paint(painter, option, index)
        value = index.data(self._role)
        if not value:
            return
        mode = gui.Icon.Mode.Normal

        if not option.state & widgets.Style.StateFlag.State_Enabled:
            mode = gui.Icon.Mode.Disabled
        elif option.state & widgets.Style.StateFlag.State_Selected:
            mode = gui.Icon.Mode.Selected
        match value:
            case gui.QPixmap():
                icon = gui.QIcon(value)
                option.decorationSize = int(value.size() / value.devicePixelRatio())

            case gui.QColor():
                pixmap = gui.QPixmap(option.decorationSize)
                pixmap.fill(value)
                icon = gui.QIcon(pixmap)

            case gui.QImage():
                icon = gui.QIcon(gui.QPixmap.fromImage(value))
                option.decorationSize = int(value.size() / value.devicePixelRatio())

            case gui.QIcon():
                icon = value
                is_on = option.state & widgets.Style.StateFlag.State_Open
                state = gui.Icon.State.On if is_on else gui.Icon.State.Off
                actual_size = option.icon.actualSize(option.decorationSize, mode, state)
                option.decorationSize.boundedTo(actual_size)
            case _:
                raise ValueError(value)
        r = core.Rect(core.Point(), option.decorationSize)
        r.moveCenter(option.rect.center())
        r.setRight(option.rect.right() - self.margin)
        state = (
            gui.Icon.State.On
            if option.state & widgets.Style.StateFlag.State_Open
            else gui.Icon.State.Off
        )
        alignment = constants.ALIGN_RIGHT | constants.ALIGN_V_CENTER
        icon.paint(painter, r, alignment, mode, state)


if __name__ == "__main__":
    from prettyqt import iconprovider

    app = widgets.app()
    model = gui.StandardItemModel()
    model.add("test")
    app = widgets.app()
    w = widgets.ListView()
    w.set_delegate(IconDelegate())
    w.set_model(model)
    item = gui.StandardItem("Item")
    item.set_data(iconprovider.get_icon("mdi.folder"), "user")
    model += item
    w.show()
    app.exec()



from __future__ import annotations

import contextlib
from itertools import takewhile
import logging
from typing import Literal

from pygments import highlight, lexers, styles
from pygments.formatter import Formatter

from prettyqt import core, gui, paths
from prettyqt.utils import get_repr


# inspired by  https://github.com/Vector35/snippets/blob/master/QCodeEditor.py
# (MIT license) and
# https://pygments.org/docs/formatterdevelopment/#html-3-2-formatter

logger = logging.getLogger(__name__)


StyleStr = Literal[
    "default",
    "emacs",
    "friendly",
    "friendly_grayscale",
    "colorful",
    "autumn",
    "murphy",
    "manni",
    "material",
    "monokai",
    "perldoc",
    "pastie",
    "borland",
    "trac",
    "native",
    "fruity",
    "bw",
    "vim",
    "vs",
    "tango",
    "rrt",
    "xcode",
    "igor",
    "paraiso-light",
    "paraiso-dark",
    "lovelace",
    "algol",
    "algol_nu",
    "arduino",
    "rainbow_dash",
    "abap",
    "solarized-dark",
    "solarized-light",
    "sas",
    "staroffice",
    "stata",
    "stata-light",
    "stata-dark",
    "inkpot",
    "zenburn",
    "gruvbox-dark",
    "gruvbox-light",
    "dracula",
    "one-dark",
    "lilypond",
    "nord",
    "nord-darker",
    "github-dark",
]


class PygmentsHighlighter(gui.SyntaxHighlighter):
    """Syntax highlighter that uses Pygments for parsing."""

    # ---------------------------------------------------------------------------
    #  "QSyntaxHighlighter" interface
    # ---------------------------------------------------------------------------

    def __init__(
        self,
        parent: gui.QTextDocument,
        lexer: str,
        style: None | StyleStr = None,
    ):
        super().__init__(parent)

        self._document = self.document()
        self._style = None
        self._stylename = ""
        if style is None:
            gui.GuiApplication.styleHints().colorSchemeChanged.connect(
                self.adjust_style_to_palette
            )
        self.set_style(style)
        if lexer == "regex":
            self._lexer = lexers.load_lexer_from_file(str(paths.RE_LEXER_PATH))
        else:
            self._lexer = lexers.get_lexer_by_name(lexer)
        self._formatter = QFormatter(style=self._stylename)

    @property
    def background_color(self):
        return self._formatter.style.background_color

    def highlightBlock(self, text):
        cb = self.currentBlock()
        p = cb.position()
        text_ = self.document().toPlainText() + "\n"
        highlight(text_, self._lexer, self._formatter)

        enters = sum(1 for _ in takewhile(lambda x: x == "\n", text_))
        # pygments lexer ignore leading empty lines, so we need to do correction
        # here calculating the number of empty lines.

        # dirty, dirty hack
        # The core problem is that pygemnts by default use string streams,
        # that will not handle QTextCharFormat, so we need use `data` property to
        # work around this.
        for i in range(len(text)):
            with contextlib.suppress(IndexError):
                self.setFormat(i, 1, self._formatter.data[p + i - enters])

    def __repr__(self):
        return get_repr(self, lexer=self._lexer.aliases[0])

    def set_style(self, style: None | StyleStr):
        match style:
            case None:
                self.adjust_style_to_palette()
                return
            case str():
                self._style = styles.get_style_by_name(style)
                self._stylename = style
            case _:
                raise TypeError(style)
        self._formatter = QFormatter(style=self._stylename)

    def get_style(self) -> StyleStr:
        return self._stylename

    def adjust_style_to_palette(self):
        pal = gui.GuiApplication.get_palette()
        style = "monokai" if pal.is_dark() else "default"
        self.set_style(style)

    style = core.Property(
        str,
        get_style,
        set_style,
        doc="Pygments style for the highlighter",
    )


def get_text_char_format(style):
    text_char_format = gui.QTextCharFormat()
    if hasattr(text_char_format, "setFontFamilies"):
        text_char_format.setFontFamilies(["monospace"])
    else:
        text_char_format.setFontFamily("monospace")
    if style.get("color"):
        text_char_format.setForeground(gui.QColor(f"#{style['color']}"))

    if style.get("bgcolor"):
        text_char_format.setBackground(gui.QColor(style["bgcolor"]))

    if style.get("bold"):
        text_char_format.setFontWeight(gui.QFont.Weight.Bold)
    if style.get("italic"):
        text_char_format.setFontItalic(True)
    if style.get("underline"):
        text_char_format.setFontUnderline(True)

    # TODO find if it is possible to support border style.

    return text_char_format


class QFormatter(Formatter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.data = []
        self._style = {name: get_text_char_format(style) for name, style in self.style}

    def format(self, tokensource, outfile):
        """Format the given token stream.

        `outfile` is argument from parent class, but
        in Qt we do not produce string output, but QTextCharFormat, so it needs to be
        collected using `self.data`.
        """
        self.data = []

        for token, value in tokensource:
            self.data.extend([self._style[token]] * len(value))


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    app.set_style("Fusion")
    editor = widgets.PlainTextEdit()
    highlighter = PygmentsHighlighter(editor.document(), lexer="python")
    # highlighter.set_style("monokai")
    editor.show()
    app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.qt import QtStateMachine
from prettyqt.utils import bidict


TransitionTypeStr = Literal["exclusive", "parallel"]

TRANSITION_TYPE: bidict[
    TransitionTypeStr, QtStateMachine.QAbstractTransition.TransitionType
] = bidict(
    exclusive=QtStateMachine.QAbstractTransition.TransitionType.ExternalTransition,
    parallel=QtStateMachine.QAbstractTransition.TransitionType.InternalTransition,
)


class AbstractTransitionMixin(core.ObjectMixin):
    def set_transition_type(
        self, typ: TransitionTypeStr | QtStateMachine.QAbstractTransition.TransitionType
    ):
        """Set transition type.

        Args:
            typ: transition type to use
        """
        self.setTransitionType(TRANSITION_TYPE.get_enum_value(typ))

    def get_transition_type(self) -> TransitionTypeStr:
        """Return current transition type.

        Returns:
            transition type
        """
        return TRANSITION_TYPE.inverse[self.transitionType()]


class AbstractTransition(AbstractTransitionMixin, QtStateMachine.QAbstractTransition):
    pass



from __future__ import annotations

from prettyqt import core, gui
from prettyqt.utils import get_repr


class LinearGradient(gui.GradientMixin, gui.QLinearGradient):
    """Used in combination with QBrush to specify a linear gradient brush."""

    def __repr__(self):
        return get_repr(self, self.get_start(), self.get_final_stop())

    def get_start(self) -> core.PointF:
        return core.PointF(self.start())

    def get_final_stop(self) -> core.PointF:
        return core.PointF(self.finalStop())

    def get_css(self) -> str:
        """Convert gradient to a CSS string. Can be used for stylesheets."""
        stop, final_stop = self.start(), self.finalStop()
        x1, y1, x2, y2 = stop.x(), stop.y(), final_stop.x(), final_stop.y()
        stops = self.stops()
        stops = "\n".join(f"    stop: {stop:f} {color.name()}" for stop, color in stops)
        spread = self.get_spread()
        return (
            "qlineargradient(\n"
            f"    x1: {x1}, y1: {y1}, x2: {x2}, y2: {y2}, spread:{spread},\n"
            f"{stops})"
        )


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()

    grad = LinearGradient()
    grad.set_color_at(0, "red")
    grad.set_color_at(0.5, "blue")
    grad.set_color_at(1, "green")
    print(grad.get_css())
    btn = widgets.PushButton("TestButton")
    btn.setStyleSheet(f"QPushButton{{ background-color:{grad.get_css()} }}")
    btn.show()
    app.exec()



from __future__ import annotations

from collections.abc import Sequence
import functools
import logging
from typing import TYPE_CHECKING, Any, Literal

from prettyqt import constants, core, widgets
from prettyqt.utils import bidict, classhelpers, datatypes, listdelegators


if TYPE_CHECKING:
    from collections.abc import Generator


DelegateStr = Literal[
    "widget",
    "editor",
    "html",
    "markdown",
    "button",
    "no_focus",
    "render_link",
    "star",
    "radio",
    "progress_bar",
]

logger = logging.getLogger(__name__)

EditTriggerStr = Literal[
    "none",
    "current_changed",
    "double_clicked",
    "selected_clicked",
    "edit_key_pressed",
    "any_key_pressed",
    "all",
]

EDIT_TRIGGERS: bidict[EditTriggerStr, widgets.QAbstractItemView.EditTrigger] = bidict(
    none=widgets.QAbstractItemView.EditTrigger.NoEditTriggers,
    current_changed=widgets.QAbstractItemView.EditTrigger.CurrentChanged,
    double_clicked=widgets.QAbstractItemView.EditTrigger.DoubleClicked,
    selected_clicked=widgets.QAbstractItemView.EditTrigger.SelectedClicked,
    edit_key_pressed=widgets.QAbstractItemView.EditTrigger.EditKeyPressed,
    any_key_pressed=widgets.QAbstractItemView.EditTrigger.AnyKeyPressed,
    all=widgets.QAbstractItemView.EditTrigger.AllEditTriggers,
)

SelectionBehaviourStr = Literal["rows", "columns", "items"]

SELECTION_BEHAVIOR: bidict[
    SelectionBehaviourStr, widgets.QAbstractItemView.SelectionBehavior
] = bidict(
    rows=widgets.QAbstractItemView.SelectionBehavior.SelectRows,
    columns=widgets.QAbstractItemView.SelectionBehavior.SelectColumns,
    items=widgets.QAbstractItemView.SelectionBehavior.SelectItems,
)

SelectionModeStr = Literal["single", "extended", "multi", "none"]

SELECTION_MODE: bidict[SelectionModeStr, widgets.QAbstractItemView.SelectionMode] = (
    bidict(
        single=widgets.QAbstractItemView.SelectionMode.SingleSelection,
        extended=widgets.QAbstractItemView.SelectionMode.ExtendedSelection,
        multi=widgets.QAbstractItemView.SelectionMode.MultiSelection,
        none=widgets.QAbstractItemView.SelectionMode.NoSelection,
    )
)

ScrollModeStr = Literal["item", "pixel"]

SCROLL_MODE: bidict[ScrollModeStr, widgets.QAbstractItemView.ScrollMode] = bidict(
    item=widgets.QAbstractItemView.ScrollMode.ScrollPerItem,
    pixel=widgets.QAbstractItemView.ScrollMode.ScrollPerPixel,
)


ScrollHintStr = Literal[
    "ensure_visible", "position_at_top", "position_at_bottom", "position_at_center"
]

SCROLL_HINT: bidict[ScrollHintStr, widgets.QAbstractItemView.ScrollHint] = bidict(
    ensure_visible=widgets.QAbstractItemView.ScrollHint.EnsureVisible,
    position_at_top=widgets.QAbstractItemView.ScrollHint.PositionAtTop,
    position_at_bottom=widgets.QAbstractItemView.ScrollHint.PositionAtBottom,
    position_at_center=widgets.QAbstractItemView.ScrollHint.PositionAtCenter,
)

DragDropModeStr = Literal["none", "drag", "drop", "drag_drop", "internal_move"]

DRAG_DROP_MODE: bidict[DragDropModeStr, widgets.QAbstractItemView.DragDropMode] = bidict(
    none=widgets.QAbstractItemView.DragDropMode.NoDragDrop,
    drag=widgets.QAbstractItemView.DragDropMode.DragOnly,
    drop=widgets.QAbstractItemView.DragDropMode.DropOnly,
    drag_drop=widgets.QAbstractItemView.DragDropMode.DragDrop,
    internal_move=widgets.QAbstractItemView.DragDropMode.InternalMove,
)


DropIndicatorPositionStr = Literal["on_item", "above_item", "below_item", "on_viewport"]

DROP_INDICATOR_POSITION: bidict[
    DropIndicatorPositionStr, widgets.QAbstractItemView.DropIndicatorPosition
] = bidict(
    on_item=widgets.QAbstractItemView.DropIndicatorPosition.OnItem,
    above_item=widgets.QAbstractItemView.DropIndicatorPosition.AboveItem,
    below_item=widgets.QAbstractItemView.DropIndicatorPosition.BelowItem,
    on_viewport=widgets.QAbstractItemView.DropIndicatorPosition.OnViewport,
)


StateStr = Literal[
    "none",
    "dragging",
    "drag_selecting",
    "editing",
    "expanding",
    "collapsing",
    "animating",
]

STATE: bidict[StateStr, widgets.QAbstractItemView.State] = bidict(
    none=widgets.QAbstractItemView.State.NoState,
    dragging=widgets.QAbstractItemView.State.DraggingState,
    drag_selecting=widgets.QAbstractItemView.State.DragSelectingState,
    editing=widgets.QAbstractItemView.State.EditingState,
    expanding=widgets.QAbstractItemView.State.ExpandingState,
    collapsing=widgets.QAbstractItemView.State.CollapsingState,
    animating=widgets.QAbstractItemView.State.AnimatingState,
)


class AbstractItemViewMixin(widgets.AbstractScrollAreaMixin):
    model_changed = core.Signal(core.QAbstractItemModel)

    def __init__(
        self,
        *args,
        horizontal_scroll_mode="pixel",
        vertical_scroll_mode="pixel",
        **kwargs,
    ):
        super().__init__(
            *args,
            horizontal_scroll_mode=horizontal_scroll_mode,
            vertical_scroll_mode=vertical_scroll_mode,
            **kwargs,
        )
        from prettyqt.utils import proxifier

        self.proxifier = proxifier.Proxifier(self)

    def __len__(self) -> int:
        return model.rowCount() if (model := self.model()) is not None else 0

    def _get_map(self):
        maps = super()._get_map()
        maps |= {
            "dragDropMode": DRAG_DROP_MODE,
            "horizontalScrollMode": SCROLL_MODE,
            "verticalScrollMode": SCROLL_MODE,
            "selectionMode": SELECTION_MODE,
            "selectionBehavior": SELECTION_BEHAVIOR,
            "defaultDropAction": constants.DROP_ACTION,
            "textElideMode": constants.TEXT_ELIDE_MODE,
            "editTriggers": EDIT_TRIGGERS,
        }
        return maps

    def selectAll(self):
        """Override, we dont want to selectAll for too many items bc of performance."""
        if self.model() is None:
            return
        if self.model().rowCount() * self.model().columnCount() > 1_000_000:  # noqa: PLR2004
            logger.warning("Too many cells to select.")
            return
        super().selectAll()

    def set_model(self, model: core.QAbstractItemModel | None):
        """Set the model of this View."""
        # Delete old selection model explicitely, seems to help with memory usage.
        old_model = self.model()
        old_sel_model = self.selectionModel()
        if old_model is not None or model is not None:
            self.setModel(model)
            if model is not None:
                # always set a parent. PySide6 needs one for proxying.
                model.setParent(self)
                self.setSelectionModel(core.ItemSelectionModel(model))
            self.model_changed.emit(model)
        # if old_model:
        #     old_model.deleteLater()
        #     del old_model
        if old_sel_model:
            old_sel_model.deleteLater()
            del old_sel_model
        return model

    def set_model_for(self, data: Any):
        """Set model for given data type.

        Pass any data structure and an appropriate model will be chosen automatically.

        Args:
            data: data to choose model for.
        """
        # we import to collect the models
        from prettyqt import itemmodels  # noqa: F401

        for kls in classhelpers.get_subclasses(core.QAbstractItemModel):
            if (
                "supports" in kls.__dict__
                and callable(kls.supports)
                and kls.supports(data)
                and kls.__name__ != "PythonObjectTreeModel"
            ):
                logger.debug("found model for data structure %r", data)
                break
        else:
            msg = "No suiting model found."
            raise TypeError(msg)
        model = kls(data, parent=self)
        self.set_model(model)

    def get_model(self, skip_proxies: bool = False) -> core.QAbstractItemModel:
        """Get current model of the ItemView.

        Arguments:
            skip_proxies: Whether to get current model or the non-proxy sourceModel.
        """
        model = self.model()
        if skip_proxies:
            while isinstance(model, core.QAbstractProxyModel):
                model = model.sourceModel()
        return model

    def get_models(
        self, proxies_only: bool = False
    ) -> listdelegators.ListDelegator[core.QAbstractProxyModel]:
        """Get a list of all (proxy) models connected to this view.

        Arguments:
            proxies_only: whether the non-proxy sourceModel should be included.
        """
        model = self.model()
        models = []
        while isinstance(model, core.QAbstractProxyModel):
            models.append(model)
            model = model.sourceModel()
        if (not proxies_only) and model is not None:
            models.append(model)
        return listdelegators.ListDelegator(models)

    def set_current_index(
        self,
        index: core.QModelIndex | tuple | None,
        operation: Literal["select", "deselect", "toggle"] = "select",
        *,
        clear: bool = True,
        current: bool = False,
        expand: Literal["rows", "columns"] | None = None,
    ):
        """Set current index.

        Arguments:
            index: Index to set.
            operation: Whether to select, deselect or toggle the current state.
            clear: Clear Whether to clear previously selected indexes.
            current: Current selection will be updated.
            expand: Whether to expand selection to whole column / row.
        """
        match index:
            case None:
                self.selectionModel().setCurrentIndex(
                    index, core.ItemSelectionModel.SelectionFlag.Clear
                )
                return
            case tuple():
                index = self.model().index(*index)
            case core.QModelIndex():
                pass
            case _:
                raise ValueError(index)
        match operation:
            case "select":
                flag = core.ItemSelectionModel.SelectionFlag.Select
            case "deselect":
                flag = core.ItemSelectionModel.SelectionFlag.Deselect
            case "toggle":
                flag = core.ItemSelectionModel.SelectionFlag.Toggle
            case _:
                raise ValueError(operation)
        if clear:
            flag |= core.ItemSelectionModel.SelectionFlag.Clear
        if current:
            flag |= core.ItemSelectionModel.SelectionFlag.Current
        match expand:
            case "rows":
                flag |= core.ItemSelectionModel.SelectionFlag.Rows
            case "columns":
                flag |= core.ItemSelectionModel.SelectionFlag.Columns
            case None:
                pass
            case _:
                raise ValueError(expand)
        self.selectionModel().setCurrentIndex(index, flag)

    def select_index(
        self,
        index: core.QModelIndex | tuple | None,
        operation: Literal["select", "deselect", "toggle"] = "select",
        *,
        clear: bool = True,
        current: bool = False,
        expand: Literal["rows", "columns"] | None = None,
    ):
        # index = self.model().index(self._selected_index)
        match index:
            case None:
                self.selectionModel().setCurrentIndex(
                    core.ModelIndex(), core.ItemSelectionModel.SelectionFlag.Clear
                )
                return
            case tuple():
                index = self.model().index(*index)
            case core.QModelIndex():
                pass
            case _:
                raise ValueError(index)
        match operation:
            case "select":
                flag = core.ItemSelectionModel.SelectionFlag.Select
            case "deselect":
                flag = core.ItemSelectionModel.SelectionFlag.Deselect
            case "toggle":
                flag = core.ItemSelectionModel.SelectionFlag.Toggle
            case _:
                raise ValueError(operation)
        if clear:
            flag |= core.ItemSelectionModel.SelectionFlag.Clear
        if current:
            flag |= core.ItemSelectionModel.SelectionFlag.Current
        match expand:
            case "rows":
                flag |= core.ItemSelectionModel.SelectionFlag.Rows
            case "columns":
                flag |= core.ItemSelectionModel.SelectionFlag.Columns
            case None:
                pass
            case _:
                raise ValueError(expand)
        self.selectionModel().select(index, flag)

    # def move_selection(self, row_offset: int = 0, column_offset: int = 0):
    #     """Move selection by an offset."""
    #     for index in self.selectedIndexes():
    #         new_idx = self.model().index(
    #             index.row() + row_offset, index.column() + column_offset, index.parent()
    #         )
    #         if new_idx.isValid():
    #             self.set_current_index(new_idx, current=True, expand="rows")

    def move_row_selection(self, dy: int):
        for row in self.selected_rows():
            new_idx = self.model().index(row + dy, 0)
            if new_idx.isValid():
                self.set_current_index(new_idx, current=True, expand="rows")

    def set_delegate(
        self,
        delegate: widgets.QAbstractItemDelegate | DelegateStr | None,
        *,
        column: int | Sequence | None = None,
        row: int | Sequence | None = None,
        persistent: bool = False,
        **kwargs: Any,
    ):
        """Set a item delegate for the view.

        Arguments:
            delegate: Delegate to set. Can also be the id of the delegate.
            column: Column the delegate should be set for.
            row: Row the column should be set for.
            persistent: If True, open persistent editors for given area.
            kwargs: Keyword args to pass to the Delegate ctor if delegate is set by id.
        """
        match delegate:
            case widgets.QAbstractItemDelegate():
                dlg = delegate
            # case "editor":
            #     delegate = itemdelegates.EditorDelegate(parent=self, **kwargs)
            # case "widget":
            #     delegate = itemdelegates.WidgetDelegate(parent=self, **kwargs)
            # case "html":
            #     delegate = itemdelegates.HtmlItemDelegate(parent=self, **kwargs)
            # case "button":
            #     delegate = itemdelegates.ButtonDelegate(parent=self, **kwargs)
            case str():
                kls = classhelpers.get_class_for_id(widgets.StyledItemDelegate, delegate)
                dlg = kls(parent=self, **kwargs)
            case None:
                dlg = widgets.StyledItemDelegate()
            case _:
                raise ValueError(delegate)
        match column, row:
            case int(), int():
                msg = "Only set column or row, not both."
                raise ValueError(msg)
            case Sequence(), None:
                for i in column:
                    self.set_delegate(delegate, column=i, row=row, persistent=persistent)
            case int(), None:
                self.setItemDelegateForColumn(column, dlg)
                if persistent:
                    model = self.model()
                    for i in range(model.rowCount()):
                        index = model.index(i, column)
                        self.openPersistentEditor(index)
            case None, Sequence():
                for i in row:
                    self.set_delegate(
                        delegate, column=column, row=i, persistent=persistent
                    )
            case None, int():
                self.setItemDelegateForRow(row, dlg)
                if persistent:
                    model = self.model()
                    for i in range(model.columnCount()):
                        self.openPersistentEditor(model.index(row, i))
            case None, None:
                self.setItemDelegate(dlg)
                if persistent:
                    model = self.model()
                    for i in range(model.rowCount()):
                        for j in range(model.columnCount()):
                            self.openPersistentEditor(model.index(i, j))
        return dlg

    def toggle_select_all(self):
        """Select all items from list (deselect when all selected)."""
        if self.selectionModel() is None:
            return
        if self.selectionModel().hasSelection():
            self.clearSelection()
        else:
            self.selectAll()

    def set_table_color(self, color: str):
        with self.edit_stylesheet() as ss:
            ss.QHeaderView.section.backgroundColor.setValue(color)

    def current_index(self) -> core.QModelIndex | None:
        if (model := self.selectionModel()) is not None:
            return model.currentIndex()
        return None

    def current_data(self, role=constants.USER_ROLE):
        if (model := self.selectionModel()) is not None:
            idx = model.currentIndex()
            return idx.data(role)
        return None

    def current_row(self) -> int | None:
        if (model := self.selectionModel()) is not None:
            return model.currentIndex().row()
        return None

    def current_column(self) -> int | None:
        if (model := self.selectionModel()) is not None:
            return model.currentIndex().column()
        return None

    def selected_indexes(self) -> listdelegators.ListDelegator[core.QModelIndex]:
        """Return list of selected indexes in first row."""
        indexes = (x for x in self.selectedIndexes() if x.column() == 0)
        indexes = sorted(indexes, key=lambda x: x.row())
        return listdelegators.ListDelegator(indexes)

    def selected_names(self) -> Generator[Any, None, None]:
        """Return generator yielding item names."""
        return (x.data(constants.NAME_ROLE) for x in self.selected_indexes())

    def selected_rows(self) -> Generator[int, None, None]:
        """Return generator yielding row nums."""
        return (x.row() for x in self.selected_indexes())

    def selected_data(
        self, role: constants.ItemDataRole = constants.USER_ROLE
    ) -> Generator[Any, None, None]:
        """Return generator yielding selected userData.

        Arguments:
            role: ItemRole to return data for.
        """
        return (x.data(role) for x in self.selected_indexes())

    def setup_dragdrop_move(self):
        self.setDragEnabled(True)
        self.setAcceptDrops(True)
        self.setDragDropMode(self.DragDropMode.DragDrop)
        self.setDefaultDropAction(constants.MOVE_ACTION)
        self.setDropIndicatorShown(True)

    def set_edit_triggers(self, *triggers: EditTriggerStr | None):
        items = ["none" if t is None else t for t in triggers]
        flags = EDIT_TRIGGERS.merge_flags(items)
        self.setEditTriggers(flags)

    def get_edit_triggers(self) -> list[EditTriggerStr]:
        return EDIT_TRIGGERS.get_list(self.editTriggers())

    def set_selection_behavior(
        self,
        behavior: SelectionBehaviourStr | widgets.QAbstractItemView.SelectionBehavior,
    ):
        """Set selection behavior for given item view.

        Arguments:
            behavior: Selection behavior
        """
        self.setSelectionBehavior(SELECTION_BEHAVIOR.get_enum_value(behavior))

    def get_selection_behavior(self) -> SelectionBehaviourStr:
        """Return current selection behavior.

        Returns:
            selection behavior
        """
        return SELECTION_BEHAVIOR.inverse[self.selectionBehavior()]

    def get_drop_indicator_position(self) -> DropIndicatorPositionStr:
        """Return position of the drop indicator in relation to the closest item."""
        return DROP_INDICATOR_POSITION.inverse[self.dropIndicatorPosition()]

    def set_drag_drop_mode(
        self, mode: DragDropModeStr | widgets.QAbstractItemView.DragDropMode
    ):
        """Set drag-drop mode for given item view.

        Args:
            mode: drag-drop mode to use
        """
        self.setDragDropMode(DRAG_DROP_MODE.get_enum_value(mode))

    def get_drag_drop_mode(self) -> DragDropModeStr:
        """Return current drag-drop mode."""
        return DRAG_DROP_MODE.inverse[self.dragDropMode()]

    def set_state(self, state: StateStr | widgets.QAbstractItemView.State):
        """Set state for given item view.

        Args:
            state: state to use
        """
        self.setState(STATE.get_enum_value(state))

    def get_state(self) -> StateStr:
        """Return current state."""
        return STATE.inverse[self.state()]

    def set_selection_mode(
        self, mode: SelectionModeStr | widgets.QAbstractItemView.SelectionMode | None
    ):
        """Set selection mode for given item view.

        Args:
            mode: selection mode to use
        """
        if mode is None:
            mode = "none"
        self.setSelectionMode(SELECTION_MODE.get_enum_value(mode))

    def get_selection_mode(self) -> SelectionModeStr:
        """Return current selection mode."""
        return SELECTION_MODE.inverse[self.selectionMode()]

    def set_scroll_mode(self, mode: ScrollModeStr | widgets.QAbstractItemView.ScrollMode):
        """Set the scroll mode for both directions.

        Args:
            mode: mode to set
        """
        self.setHorizontalScrollMode(SCROLL_MODE.get_enum_value(mode))
        self.setVerticalScrollMode(SCROLL_MODE.get_enum_value(mode))

    def set_horizontal_scroll_mode(
        self, mode: ScrollModeStr | widgets.QAbstractItemView.ScrollMode
    ):
        """Set the horizontal scroll mode.

        Args:
            mode: mode to set
        """
        self.setHorizontalScrollMode(SCROLL_MODE.get_enum_value(mode))

    def get_horizontal_scroll_mode(self) -> ScrollModeStr:
        """Return current horizontal scroll mode."""
        return SCROLL_MODE.inverse[self.horizontalScrollMode()]

    def set_vertical_scroll_mode(
        self, mode: ScrollModeStr | widgets.QAbstractItemView.ScrollMode
    ):
        """Set the vertical scroll mode.

        Args:
            mode: mode to set
        """
        self.setVerticalScrollMode(SCROLL_MODE.get_enum_value(mode))

    def get_vertical_scroll_mode(self) -> ScrollModeStr:
        """Return current vertical scroll mode."""
        return SCROLL_MODE.inverse[self.verticalScrollMode()]

    def num_selected(self) -> int:
        """Return amount of selected rows.

        Returns:
            amount of selected rows
        """
        if (model := self.selectionModel()) is not None:
            return len(model.selectedRows())
        return 0

    def jump_to_column(self, col_num: int):
        """Make sure column at given index is visible.

        scrolls to column at given index

        Args:
            col_num: column to scroll to
        """
        if (model := self.model()) is not None:
            idx = model.index(0, col_num)
            self.scrollTo(idx)

    def scroll_to_top(self):
        """Override to use abstractitemview-way of scrolling to top."""
        self.scrollToTop()

    def scroll_to_bottom(self):
        """Override to use abstractitemview-way of scrolling to bottom."""
        self.scrollToBottom()

    def select_last_row(self):
        idx = self.model().createIndex(self.model().rowCount() - 1, 0)
        self.setCurrentIndex(idx)

    def select_first_row(self):
        idx = self.model().index(0, 0)
        self.set_current_index(idx, current=True, expand="rows")

    def scroll_to(
        self,
        index: core.QModelIndex,
        mode: ScrollHintStr | widgets.QAbstractItemView.ScrollHint = "ensure_visible",
    ):
        self.scrollTo(index, SCROLL_HINT.get_enum_value(mode))

    def highlight_when_inactive(self):
        """Highlight items when widget does not have focus."""
        p = self.get_palette()
        p.highlight_inactive()
        self.setPalette(p)

    def set_icon_size(self, size: datatypes.SizeType):
        self.setIconSize(datatypes.to_size(size))

    def get_size_hint_for_column(self, col: int, row_limit: int = 25) -> int:
        """Get a size hint for given column by finding widest cell.

        Arguments:
            col: columnt to get size hint for.
            row_limit: number of rows to check.
        """
        to_check = min(row_limit, self.model().rowCount())
        return max(
            self.sizeHintForIndex(self.model().index(row, col)).width()
            for row in range(to_check)
        )

    def sync_with(
        self,
        table_to_sync: widgets.QAbstractItemView,
        orientation: constants.OrientationStr | constants.Orientation,
    ) -> list[core.QMetaObject.Connection]:
        orientation = constants.ORIENTATION.get_enum_value(orientation)
        """Sync ItemView cell widths and scrollbars with another ItemView.

        Arguments:
            table_to_sync: Table to sync with
            orientation: Whether to sync horizontal or vertical orientation.
        """

        def _table_resized(col, _, new_size, table, orientation):
            if orientation == constants.HORIZONTAL:
                table.setColumnWidth(col, new_size)
            else:
                table.setRowHeight(col, new_size)

        _table_1_resized = functools.partial(
            _table_resized, table=self, orientation=orientation
        )
        _table_2_resized = functools.partial(
            _table_resized, table=table_to_sync, orientation=orientation
        )
        if orientation == constants.VERTICAL:
            h1 = self.v_scrollbar.valueChanged.connect(table_to_sync.v_scrollbar.setValue)
            h2 = table_to_sync.v_scrollbar.valueChanged.connect(self.v_scrollbar.setValue)
            h3 = self.v_header.sectionResized.connect(_table_2_resized)
            h4 = table_to_sync.v_header.sectionResized.connect(_table_1_resized)
        else:
            h1 = self.h_scrollbar.valueChanged.connect(table_to_sync.h_scrollbar.setValue)
            h2 = table_to_sync.h_scrollbar.valueChanged.connect(self.h_scrollbar.setValue)
            h3 = self.h_header.sectionResized.connect(_table_2_resized)
            h4 = table_to_sync.h_header.sectionResized.connect(_table_1_resized)
        return [h1, h2, h3, h4]

    def get_visible_section_span(
        self, orientation: constants.OrientationStr | constants.Orientation
    ) -> tuple[int, int]:
        """Get range of visible sections.

        Arguments:
            orientation: Whether to get section span for horizontal or vertical header.
        """
        orientation = constants.ORIENTATION.get_enum_value(orientation)
        top_left = core.QPoint(0, 0)
        bottom_right = self.viewport().rect().bottomRight()
        if orientation == constants.HORIZONTAL:
            start = self.indexAt(top_left).column()
            count = self.model().columnCount()
            end = self.indexAt(bottom_right).column()
        else:
            start = self.indexAt(top_left).row()
            count = self.model().rowCount()
            end = self.indexAt(bottom_right).row()
        if count == 0:
            return (-1, -1)
        end = count if end == -1 else end + 1
        return (start, end)


class AbstractItemView(AbstractItemViewMixin, widgets.QAbstractItemView):
    """The basic functionality for item view classes."""


if __name__ == "__main__":
    import pandas as pd

    df = pd.DataFrame(dict(a=[1]))
    app = widgets.app()
    table = widgets.TableView()
    table.set_model_for(df)
    table.show()
    app.exec()



from __future__ import annotations

from prettyqt import constants, widgets
from prettyqt.utils import get_repr


class GroupBox(widgets.WidgetMixin, widgets.QGroupBox):
    """Group box frame with a title.

    A group box provides a frame, a title on top, a keyboard shortcut,
    and displays various other widgets inside itself.
    The keyboard shortcut moves keyboard focus to one of the group box's child widgets.
    """

    def __repr__(self):
        return get_repr(self, self.title())

    def set_title(self, title: str):
        self.setTitle(title)

    def set_alignment(
        self, alignment: constants.HorizontalAlignmentStr | constants.AlignmentFlag
    ):
        """Set the title alignment of the groupbox.

        Args:
            alignment: title alignment for the groupbox
        """
        self.setAlignment(constants.H_ALIGNMENT.get_enum_value(alignment))

    def get_alignment(self) -> constants.HorizontalAlignmentStr:
        """Return current title alignment.

        Returns:
            title alignment
        """
        return constants.H_ALIGNMENT.inverse[self.alignment()]

    def set_enabled(self, state):
        for widget in self.layout():
            widget.setEnabled(state)


if __name__ == "__main__":
    app = widgets.app()
    widget = GroupBox()
    ly = widgets.HBoxLayout()
    ly += widgets.RadioButton("test")
    widget.set_layout(ly)
    widget.show()
    app.exec()



"""Top-level package for PrettyQt."""

from importlib.metadata import version as _metadata_version
import importlib
import logging
import inspect

__author__ = """Philipp Temminghoff"""
__email__ = "phil65@kodi.tv"
__version__ = _metadata_version("prettyqt")

# import os

# os.environ["PYTHONBREAKPOINT"] = "prettyqt.debug"


def import_all():
    from prettyqt import qt

    logger = logging.getLogger(__name__)
    mods = []
    for mod in __all__:
        try:
            mod = importlib.import_module(f"prettyqt.{mod}")
            mods.append(mod)
        except ImportError:
            logger.warning("%s not available. binding: %s", mod, qt.API)
    return mods


def debug():
    """Print the local variables in the caller's frame."""
    frame = inspect.currentframe()
    stack = inspect.stack()
    if frame is None or frame.f_back is None:
        del frame
        raise RuntimeError
    try:
        from prettyqt import custom_widgets, widgets

        debug.app = widgets.app(organization_name="phil65", application_name="Prettyqt")
        objectbrowser = custom_widgets.ObjectBrowser(frame.f_back.f_globals, stack)
        objectbrowser.show()
        debug.app.exec()
    finally:
        del frame


def show(item):
    from prettyqt import core, debugging, widgets, itemmodels  # noqa
    from prettyqt.utils import classhelpers

    debug.app = widgets.app(organization_name="phil65", application_name="Prettyqt")
    for klass in classhelpers.get_subclasses(core.QAbstractItemModel):
        if "supports" in klass.__dict__ and klass.supports(item):
            if issubclass(klass, core.QAbstractTableModel):
                viewer = widgets.TableView()
            else:
                viewer = widgets.TreeView()
            instance = klass(item)
            viewer.set_model(instance)
            viewer.show()
            debug.app.exec()
            break
    debug.app.exec()


__all__ = [
    "animations",
    "bluetooth",
    "charts",
    "constants",
    "core",
    "custom_network",
    "custom_widgets",
    "debugging",
    "designer",
    "docs",
    "eventfilters",
    "gui",
    "iconprovider",
    "ipython",
    "itemdelegates",
    "itemmodels",
    "location",
    "multimedia",
    "multimediawidgets",
    "network",
    "openglwidgets",
    "pdf",
    "pdfwidgets",
    "positioning",
    "prettyqtmarkdown",
    "printsupport",
    "qml",
    "qthelp",
    "quick",
    "quickwidgets",
    "scintilla",
    "scxml",
    "spatialaudio",
    "statemachine",
    "svg",
    "svgwidgets",
    "syntaxhighlighters",
    "texttospeech",
    "utils",
    "validators",
    "webchannel",
    "webenginecore",
    "webenginewidgets",
    "widgets",
]


if __name__ == "__main__":
    from importlib import metadata

    dist = metadata.distribution("prettyqt")
    show(["abc"])



from __future__ import annotations

from prettyqt import designer


class FormBuilder(
    designer.abstractformbuilder.AbstractFormBuilderMixin, designer.QFormBuilder
):
    """Used to dynamically construct user interfaces from UI files at run-time."""


if __name__ == "__main__":
    builder = FormBuilder()



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import core, widgets


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


class SizeEdit(widgets.Widget):
    value_changed = core.Signal(core.Size)

    def __init__(self, *args, object_name: str = "size_edit", **kwargs):
        super().__init__(*args, object_name=object_name, **kwargs)
        self.set_margin(0)
        self.spinbox_width = widgets.SpinBox(maximum=999999)
        self.spinbox_height = widgets.SpinBox(maximum=999999)
        layout = self.set_layout("horizontal", margin=0)
        layout.add(widgets.Label("width", alignment="center_right"))
        layout.add(self.spinbox_width)
        layout.add(widgets.Label("height", alignment="center_right"))
        layout.add(self.spinbox_height)
        self.spinbox_width.value_changed.connect(self._on_value_change)
        self.spinbox_height.value_changed.connect(self._on_value_change)

    def _on_value_change(self):
        self._value = self.get_value()
        self.value_changed.emit(self._value)

    def get_value(self) -> core.Size:
        return core.Size(
            self.spinbox_width.get_value(),
            self.spinbox_height.get_value(),
        )

    def set_value(self, value: datatypes.SizeType):
        if isinstance(value, tuple):
            value = core.Size(*value)
        self._value = value
        self.spinbox_width.set_value(value.width())
        self.spinbox_height.set_value(value.height())

    value = core.Property(
        core.Size,
        get_value,
        set_value,
        user=True,
        doc="Current value",
    )


if __name__ == "__main__":
    app = widgets.app()
    widget = SizeEdit(window_title="Test")
    widget.value_changed.connect(print)
    widget.show()
    app.exec()
